// $Id: NUS_Group.cpp,v 1.1 2006/08/10 17:16:01 nicholas Exp $


#include "NUS_Group.h"
#include "NUS_Order.h"
#include "NUS_Player.h"

#include <vector>
#include <assert.h>
#include <math.h>

static const bool debug = false;

namespace NUS_Simul{

	Group::Group(uint4 playerID) : MapObject(playerID){

		currentOrder = NULL;
		maxSpeed = 0.0;
		maxAttackRange = 0.0;
		lastDistanceMoved = 0.0;

		radius = 0.0;

	}

    Group::Group(uint4 playerID, UnitVector &unitVector) : MapObject(playerID) {
		currentOrder = NULL;
		maxSpeed = 0.0;
		maxAttackRange = 0.0;
		lastDistanceMoved = 0.0;

		radius = 0.0;

		units = unitVector;
		
		recalculateRadius();

		for (Vector<Unit*>::iterator it = units.begin(); it != units.end(); it++) {
			if ((*it)->getObject()->get_int("max_speed") > maxSpeed)
				maxSpeed = (*it)->getObject()->get_int("max_speed");
			if ((*it)->getObject()->component("weapon")->get_int("max_ground_range") > maxAttackRange)
				maxAttackRange = (*it)->getObject()->component("weapon")->get_int("max_ground_range");
		}
	}

    Group::Group(uint4 playerID, Unit* unit) : MapObject(playerID) {
		currentOrder = NULL;
		maxSpeed = 0.0;
		maxAttackRange = 0.0;
		lastDistanceMoved = 0.0;

		radius = 0.0;

		units.push_back(unit);
		
		recalculateRadius();

		GameObj *currObj = unit->getObject();
		maxSpeed = currObj->get_int("max_speed");
		maxAttackRange = currObj->component("weapon")->get_int("max_ground_range");
	}
		
	void Group::setUnits(UnitVector &u) {
		maxSpeed = 0.0;
		maxAttackRange = 0.0;

		units = u;
		recalculateRadius();

		for (Vector<Unit*>::iterator it = units.begin(); it != units.end(); it++) {
			if ((*it)->getObject()->get_int("max_speed") > maxSpeed)
				maxSpeed = (*it)->getObject()->get_int("max_speed");
			if (((*it)->getObject()->component("weapon") != NULL) && (*it)->getObject()->component("weapon")->get_int("max_ground_range") > maxAttackRange)
				maxAttackRange = (*it)->getObject()->component("weapon")->get_int("max_ground_range");
			else
				maxAttackRange = 0;
		}
	}
	 
	Group::~Group(){
		delete currentOrder;
		currentOrder = NULL;
	}

	Group::Group(const Group& b) : MapObject(b){
	// NOTE: orders are NOT copied by default!  they must be copied manually
	// this is done because in MOST cases, we DON'T need to copy orders

		//if (debug) std::cout <<  "Group: start cloning order" << std::endl;
		
		if (b.currentOrder) 
			currentOrder = b.currentOrder->clone();
		else
			currentOrder = NULL;

		//if (debug) std::cout <<  "Group: end cloning order and start cloning units " << std::endl;
		for (Vector<Unit*>::const_iterator it = (b.units).begin(); it != (b.units).end(); it++)
			units.push_back((*it)->clone());

		//if (debug) std::cout <<  "Group: end cloning units " << std::endl;
		
		maxSpeed = b.maxSpeed;
		maxAttackRange = b.maxAttackRange;
		lastDistanceMoved = b.lastDistanceMoved;

		radius = b.radius;
		
		vec = b.vec;
	}

	Group* Group::clone() const{
		return new Group(*this);
	}

	bool Group::withinRange(Group* otherGroup) {
		return
			(position - otherGroup->getPosition()).length() <=
			(radius + otherGroup->getExtents()+otherGroup->getMaxAttackRange());
	}

	ObjectVector Group::getUnitsObject(){
		
		ObjectVector unitsObj;
		
		FORALL(units, it){
			unitsObj.push_back((*it)->getObject());
		}
		
		return unitsObj;
	}

/*

	void Group::draw(){
	// here, draw the attack ranges of all the different units in the group...
		uint4 count = 0; // used for graphics

		FORALL(unitTypes,it){
			if((it->second).numUnits > 0){
				drawCircle(position,
					(it->second).unitPrototype->getAttackRange()+radius, CIRCLE_DETAIL, transparentBlue);

	// this whole thing is really hacky... but there is no easy way
	// to make this work, and I don't want to put that much effort
	// just to make it look pretty

				if(displayObjectInfo){
					std::string unitsString = (it->second).unitPrototype->getName() +": " + convertToString((it->second).numUnits);

					if(draw3DTextInfo) {
						draw3DString(position.x-8,
							position.y-8-count*(2+cameraPos.z/-200),
							unitsString,
							color,
							1.0/(-cameraPos.z/4.0));
					}else{
						draw2DString(position.x-8,
							position.y-8-count*(2+cameraPos.z/-200),
							unitsString,
							color,
							true);
					}
				}
			}else{
				ERR2("SOMETHING TERRIBLY WRONG IN GROUP: ", name);
			}
			count++;
		}

	// draw the unit itself...
		drawCircle(position,radius,CIRCLE_DETAIL,color);
	}

	void Group::drawInSelectionMode(){
		glPushName(GROUP_TYPE_ID);

		MapObject::drawInSelectionMode();

		glPopName();
	}
*/

	void Group::setOrder(Order* newOrder){
		delete currentOrder;

		currentOrder = newOrder;
	}

	void Group::executeOrders(real8 timeElapsed){
		if(currentOrder == NULL){
			lastDistanceMoved = 0.0;
			return;
		}

		Vector3D direction = currentOrder->getDestination() - position;

		real8 adjustedMaxSpeed = currentOrder->getSpeed()*timeElapsed;

		if(direction.length() <= adjustedMaxSpeed){
	// if we have reached our current destination order...
			position = currentOrder->getDestination();
	// if we have reached our destination, then we can execute our order action...
			currentOrder->goToNextWayPoint();

		//currentOrder->setCompleted(true);
		}else{
			position+=(direction|adjustedMaxSpeed);

		}
		if(direction.length() == 0.0){
			lastDistanceMoved = 0.0;
		}else{
			lastDistanceMoved = (direction|adjustedMaxSpeed).length();
		}
	}

	void Group::executeOrderAction(SimulationState& state){
		assert(currentOrder!=NULL);

		if(currentOrder->getActionType() == "MOVE"){ // do nothing at the end, since all we were doing is moving...

		}else if(currentOrder->getActionType() == "ATTACK"){
			if(state.enemyInRange(this)){
				return; // if there are still enemies in range, don't end this order...
			}
		}

		cleanupOrder();
	}

	void Group::cleanupOrder(){
		if(currentOrder == NULL){
			ERR("Trying to cleanup an empty order, this should not happen!");
		}else{
			delete currentOrder;
			currentOrder = NULL;
		}

	}
		
	void Group::recalculateRadius() {
		// IDEAL: http://www.inf.ethz.ch/personal/gaertner/texts/own_work/esa99_final.pdf
		// current: find a bounding circle (need not be minimum)

		sint4 maxX = min_sint4;
		sint4 maxY = min_sint4;
		sint4 minX = max_sint4;
		sint4 minY = max_sint4;
		
		real8 cg_x = 0.0;
		real8 cg_y = 0.0;
		real8 totalWeights = 0.0;
		
		for (Vector<Unit*>::iterator it = units.begin(); it != units.end(); it++) {
			sint4 x,y;
			
			(*it)->getObject()->get_center(x,y);
			
			if (x > maxX) maxX = x;
			if (y > maxY) maxY = y;
			if (x < minX) minX = x;
			if (y < minY) minY = y;			
			
			cg_x += (*it)->hitPoints * x;
			cg_y += (*it)->hitPoints * y;
			totalWeights += (*it)->hitPoints;
		}
		
		//use the center of gravity for this
		//setPosition( Vector3D((maxX + minX)/2, (maxY+minY)/2, 0) );
		setPosition( Vector3D( cg_x / totalWeights, cg_y / totalWeights, 0.0) );
		
		
		radius = units[0]->getObject()->get_int("radius"); //min_sint4;
		
		real8 centerX = (maxX + minX)/2;
		real8 centerY = (maxY + minY)/2; 
				
		for (Vector<Unit*>::iterator it = units.begin(); it != units.end(); it++) {
			sint4 x,y;
			
			(*it)->getObject()->get_center(x,y);
			
			real8 dist = sqrt((centerX-x)*(centerX-x) + (centerY-y)*(centerY-y));
			
			if (dist > radius) radius = dist;
		}		
	}
	
	void Group::merge(Group *other) {
		units.insert(units.end(), other->getUnits().begin(), other->getUnits().end());
		recalculateRadius();
	}

}

