#ifndef NUS_SIMULATION_STATE_H
#define NUS_SIMULATION_STATE_H

#include <map>
//#include <unistd.h>
#include "timeval.h"

#include "NUS_TerrainBase.H"
#include "Vector.H"


#include "NUS_Base.h"

#include "NUS_Strategies.h"
#include "NUS_Player.h"
#include "NUS_Group.h"

namespace NUS_Simul{
	
enum VictoryTypes { DESTROY_ALL_BASES = 0, DESTROY_ALL_UNITS };
	
template <class T> T sgn(const T& x){
	if (x >= 0) return 1;
	return -1;
}

// get the current time (in seconds)
inline real8 getCurrentTime(){
	struct timeval tv;
	gettimeofday(&tv,NULL);
	return (tv.tv_sec + (tv.tv_usec / 1000000.0));
}

class SimulationAI;

const int MaxSearchDepth = 50;

struct SavedData{
	bool isDone;
	uint4 ourStrategyIndex;
	uint4 theirStrategyIndex;
	real8 currentValue;
	real8 bestValue;
	uint4 searchStrategyIndex[MaxSearchDepth];
	//TaskCreateFunctionPtr bestStrategy;
	uint4 bestStrategyIndex;

	SavedData(){
		isDone = false;
		ourStrategyIndex = 0;
		theirStrategyIndex = 0;
		currentValue = 0.0;
		bestValue = 0.0;
		bestStrategyIndex = 0;
	}
};

typedef SavedData SimulationResult;

typedef std::map<const uint4,Player*> PlayerMap;

typedef std::map<Group*, Player*> GroupMap;
typedef std::map<Base*, Player*> BaseMap;

class SimulationState{
public:
	SimulationState(SimulationAI* sim){
		simAI = sim;
	}
	~SimulationState();
	SimulationState(const SimulationState& b);

	SimulationState& operator= (const SimulationState& b);

	SimulationState* clone();
	void cleanup();

	SimulationAI* simAI;
	
	real4 getEnemyInfluence(uint4 playerID, NUS_TerrainBase::Loc loc);
	real4 getFriendlyInfluence(uint4 playerID, NUS_TerrainBase::Loc loc);
	
	//get nearest enemy with weaker influence
	Unit* getNearestWeakEnemyUnitOfGroup(uint4 playerID,const UnitVector& units, uint4 groupsize);

	Base* getNearestEnemyBaseOfGroup(uint4 playerID,const UnitVector& units);
	Base* getNearestEnemyBaseOfGroup(uint4 playerID,const Group *group);
	Unit* getNearestEnemyUnitOfGroup(uint4 playerID,const UnitVector& units);
	Unit* getAverageNearestEnemyUnitOfGroup(uint4 playerID,const UnitVector& units);
	Unit* getNearestEnemyUnitOfCGOfGroup(uint4 playerID,const UnitVector& units);
	Group* getNearestGroupOfGroup(uint4 playerID, const Group* group);

	NUS_TerrainBase::Loc getCenterOfObjects(const UnitVector& units);
	Unit* getNearestObject(const UnitVector& units, const NUS_TerrainBase::Loc& loc);
	Unit* getNearestObject(const UnitVector& units,Unit* obj);
	Unit* getNearestObject(const UnitVector& units,Base* obj);

	bool attackBasesInRange(Vector<Base*> &, Group *, real8);
	bool attackUnitsInRange(Vector<Group*> &, Group *, real8);

	// populates X nearest objects to a specified object
	void getNearestObjectsToObject(UnitVector& closestUnits,UnitVector originalUnits,Unit* unit,uint4 numUnits);
	void getNearestObjectsToObject(UnitVector& closestUnits,UnitVector originalUnits,Base* base,uint4 numUnits);

	bool removeUnitFromVector(Unit* unit, UnitVector& unitVector);
	bool subtractVector(const UnitVector& vectorToSubtract, UnitVector& vector);

	bool enemyInRange(Group* group);
	Player* getPlayer(uint4 playerID);

	void addPlayer(Player* player){
		if(player == NULL || getPlayer(player->getID()) != NULL){
			ERR("error in creating new player");
		}
		players[player->getID()] = player;
	}

	void setOurID(uint4 id){ ourPlayerID = id;}
	uint4 getOurID() { return ourPlayerID;}

	PlayerMap& getPlayers(){return players;}

	void updateObjects(real8 lastMoveUpdateMillis, real8 currUpdateMillis);

	TaskCreateFunctionPtr findBestStrategy(
		uint4 playerID,
		real8 currentTime,
		real8 allowedComputationTime,
		real8 timeToSimulate,
		SimulationResult& simResult);

	TaskCreateFunctionPtr searchStrategy(
		uint4 playerID,
		real8 currentTime,
		real8 allowedComputationTime,
		real8 timeToSimulate,
		SimulationResult &simResult);

	real8 negamax(
		SimulationState *currentState,
		uint4 playerID,
		uint8 depth,
		real8 currentTime,
		real8 simStartTime,
		real8 allowedComputationTime,
		real8 timeToSimulate,
		real8 alpha,
		real8 beta,
		SimulationResult &simResult);

	real8 advanceTime(real8 lastTime);

	bool isGameOver();
	bool hasWon(uint4 playerID);

	void updateStrategies();

	uint4 groupSize;
	uint4 smartJoinSize;
	bool smartJoin;
	bool useMoveForAttack;
	bool findspot;

private:
	SimulationState();

	PlayerMap players;
	uint4 ourPlayerID;

	real8 evaluate(uint4 playerID);
	real8 getNextTime(real8 currentTime);
	void updateOrders();

	void moveObjects(real8 timeElapsed);
	void resolveCollisions(real8 currentTime);
	bool solveQuadratic(real8 a,real8 b,real8 c,real8& t1,real8& t2);
	real8 getCirclePointCollisionTime(const Vector3D& circlePos,
		real8 radius,
		const Vector3D& pointPos,
		const Vector3D& relativeHeading,
		real8 startTime);

	real8 getNextCollisionTime(real8 currentTime);
	real8 getNextOrderCompletionTime(real8 currentTime);
	real8 getStrategyValue(
		uint4 playerID,
		TaskCreateFunctionPtr ourStrategy,
		TaskCreateFunctionPtr theirStrategy,
		real8 currentTime,
		real8 timeToSimulate);

	real8 simulateForTime(
		uint4 playerID,
		real8 currentTime,
		real8 timeToSimulate);

	void copyPlayerInfo(Player* origPlayer,SimulationState* tempState);
	
	GroupMap groupsInRangeMap;
	BaseMap basesInRangeMap;
};


}

REGISTER_TYPEOF(2861, NUS_Simul::PlayerMap::iterator);
REGISTER_TYPEOF(2862, NUS_Simul::PlayerMap::const_iterator);


#endif



