// $Id: NUS_Player.cpp,v 1.1 2006/08/10 17:16:02 nicholas Exp $
#include <algorithm>

#include "NUS_Group.h"
#include "NUS_Order.h"
#include "NUS_MapObject.h"
#include "NUS_Base.h"

#include "NUS_SimulationState.h"

#include "NUS_Player.h"

static const bool debug = false;

namespace NUS_Simul{

	Player::Player(uint4 playerID, sint4 width, sint4 height){
		this->playerID = playerID;
		currentStrategy = NULL;
		
		useInfluence = true;
		//influence map of attack radius
		//cell size can be better, use a tile as a cell for now
		attack_influence.init(width , height, 16);
		
		if (debug) std::cout <<"player is " << playerID <<" first value is " <<attack_influence.value_at(10,10) << std::endl;
	}
	
	Player::Player(uint4 playerID){
		this->playerID = playerID;
		currentStrategy = NULL;
		victoryType = DESTROY_ALL_BASES;
		useInfluence = false;
		
		if (debug) std::cout << "player created " << std::endl;
	}

	Player::~Player(){
	// get rid of this player's strategy
		delete currentStrategy;
		currentStrategy = NULL;

	// get rid of all the groups
		FORALL (groups, it) delete *it;
		groups.clear();

	// then get rid of all the bases
		FORALL (bases,it) delete *it;
		bases.clear();
	}


	Player::Player(const Player& p){
		playerID = p.playerID;

		//if (debug) std::cout <<  "Player: start cloning groups" << std::endl;
		FORALL(p.groups, it) {
			Group* newGroup = (*it)->clone();
			groups.push_back(newGroup);
			
			UnitVector &unitVector(newGroup->getUnits());
			
			FORALL (unitVector, uit) {
				objectUnitMap[(*uit)->getObject()] = *uit;
				unitGroupMap[*uit] = newGroup;
				
				ortsUnits[*uit] = ExtraObjectData();
			}
		}
		//if (debug) std::cout <<  "Player: done cloning groups" << std::endl;

		//if (debug) std::cout <<  "number of groups:" << groups.size() << std::endl;
		
		//if (debug) std::cout <<  "Player: start cloning bases" << std::endl;
		FORALL(p.bases, it){
			Base* newBase = (*it)->clone();
			bases.push_back(newBase);
			
			objectBaseMap[newBase->getObject()] = newBase;
			ortsBases[newBase] = ExtraObjectData();
		}
		//if (debug) std::cout <<  "Player: done cloning bases" << std::endl;

		currentStrategy = NULL;
		
		victoryType = p.victoryType;
		
		useInfluence = p.useInfluence;
	}

	Player* Player::clone()
	{
		return new Player(*this);

	}

	void Player::addGroup(Group* newGroup){
		if(newGroup == NULL){
			ERR("trying to add NULL group");
		}else{
			groups.push_back(newGroup);
			
			Vector<Unit*> &units(newGroup->getUnits());
			
			for (Vector<Unit*>::iterator uit = units.begin(); uit != units.end(); uit++) {
				if (*uit==NULL) ERR("addGroup: internal unit has address NULL");
				
				objectUnitMap[(*uit)->getObject()] = *uit;
				ortsUnits[*uit] = ExtraObjectData();
				unitGroupMap[*uit] = newGroup;
			}
		}
	}

	void Player::addBase(Base* newBase){
		if(newBase == NULL){
			ERR("trying to add NULL base");
		}else{
			bases.push_back(newBase);
		}
	}
	
	void Player::removeOrtsUnit(GameObj *obj) {
		Unit *unit = objectUnitMap[obj];
		
		removeOrtsUnit(unit);
	}
/*
	void Player::drawObjects(){
	// draw bases first, so groups will always be drawn over them
		FORALL(bases,it)  (it->second)->draw();
		FORALL(groups,it) (it->second)->draw();
	}

	void Player::drawObjectsInSelectionMode(){
	// FORALL(bases,it)  (it->second)->draw();
		FORALL(groups,it) (it->second)->drawInSelectionMode();
	}
*/

	void Player::executeOrders(real8 timeElapsed){
// OPTION 2 (See NUS_Player.H)
// if (currentStrategy) {
//	GroupVector &groups(currentStrategy->getGroups());
		FORALL(groups,it){
			Group* group = *it;
			if (group->getNumUnits() > 0){
				group->executeOrders(timeElapsed);
			}
		}
// }
	}

	void Player::updateOrders(SimulationState& state){
// OPTION 2 (See NUS_Player.H)
// if (currentStrategy) {
//	GroupVector &groups(currentStrategy->getGroups());
		FORALL(groups,it){
			Group* group = *it;

			if(group->getOrder() == NULL) continue;

			if (group->getOrder()->isCompleted()) {
				group->executeOrderAction(state);
			}
		}
	// }
	}


	void Player::setStrategy(Task* strategy){
		delete currentStrategy;

		currentStrategy = strategy;
	}

	void Player::removeGroup(Group *gp) {
		// FIXME: Time complexity is O(n), where n is length of groups list
		Vector<Group*>::iterator gp_it = find(groups.begin(), groups.end(), gp); 
		if (gp_it != groups.end()) {
			groups.erase(gp_it); 
			
			UnitVector &unitVector(gp->getUnits());
			
			for (Vector<Unit*>::iterator uit = unitVector.begin(); uit != unitVector.end(); uit++) 
				unitGroupMap.erase(*uit);
				
			delete gp;
		}
		else ERR("Player::removeGroup: Unable to find group");
	}
	
	void Player::addOrtsUnit(GameObj *obj){
		Unit *newUnit = new Unit(obj);
		
		if (newUnit==NULL) ERR("addOrtsUnit: internal unit has address NULL");
		
		// every new unit is a group upon itself
		Group *newGroup = new Group(playerID);
	
		UnitVector unitVector;
		unitVector.push_back(newUnit);
		
		newGroup->setUnits(unitVector);
		
		//drop a point on the influence map
		if(useInfluence){
			attack_influence.add_circle(newUnit->getPosition().x, newUnit->getPosition().y, 112.0, 0.5);
			attack_influence.add_circle(newUnit->getPosition().x, newUnit->getPosition().y, 224.0, 0.5);
			lastInfluencePos[newUnit] = newUnit->getPosition();
		}
		
		addGroup(newGroup);
	}
	
	void Player::doOrders(SimulationAI* simAI){
		if (currentStrategy == NULL) return;
		
		FORALL(groups, it){
			Order* currentOrder = (*it)->getOrder();
			if(currentOrder==NULL) continue;
			if(!currentOrder->isCompleted()){
				if(!currentOrder->executed){
					if(currentOrder->getActionType()=="MOVE"){
						simAI->executePath((*it));
					}else if(currentOrder->getActionType()=="ATTACK"){
						simAI->executeAttack(*it);
					}
					currentOrder->executed = true;
				}
			}
		}
	}

	void Player::removeInternalUnit(Unit *unit) {
		FORALL(groups,it){
			Group* group = *it;
			
			UnitVector &units(group->getUnits());
			
			Vector<Unit*>::iterator uit = std::find(units.begin(), units.end(), unit);
			
			if (uit != units.end()) {
				units.erase(uit);
				
				if (units.empty()) removeInternalGroup(group);
			}
		}
	}
	
	void Player::removeInternalGroup(Group *group) {
		Vector<Group*>::iterator it = std::find(groups.begin(), groups.end(), group);
		
		if (it != groups.end())
			groups.erase(it);
	}
	
	void Player::updateObject(GameObj *obj) {
		if (objectBaseMap.find(obj) != objectBaseMap.end()) {
			objectBaseMap[obj]->setHitPoints(obj->get_int("hp"));
		} else if (objectUnitMap.find(obj) != objectUnitMap.end()) {
			Unit* unit = objectUnitMap[obj];
			unit->hitPoints = obj->get_int("hp");
			
			unit->position = NUS_TerrainBase::Loc(obj->get_int("x"), obj->get_int("y"));
			
			if(useInfluence){
				//don't update if obj didn't move much
				if(unit->distance_to(lastInfluencePos[unit])>=16){
					//undo previous position on influence map
					attack_influence.add_circle(lastInfluencePos[unit].x, lastInfluencePos[unit].y, 112.0, -0.5);
					attack_influence.add_circle(lastInfluencePos[unit].x, lastInfluencePos[unit].y, 224.0, -0.5);
					
				//if(unit->distance_to(obj)){
					//if (debug) std::cout << "update unit on influence map" << std::endl;
					//undone previous position on influence map
					attack_influence.add_circle(unit->getPosition().x, unit->getPosition().y, 112.0, 0.5);
					attack_influence.add_circle(unit->getPosition().x, unit->getPosition().y, 224.0, 0.5);
		//			std::cout <<" influence updated at" << obj->get_int("x") << ", " << obj->get_int("y") << ":" \
					<< attack_influence.value_at(obj->get_int("x"),obj->get_int("y")) <<std::endl;
					lastInfluencePos[unit] = unit->getPosition();
				}
			}
			
			// update group
			unitGroupMap[objectUnitMap[obj]]->recalculateRadius();
		}		
	}

	void Player::merge(Group *original, Group *other) {
		Vector<Group*>::iterator git = std::find(groups.begin(), groups.end(), other);
		
		if (git == groups.end()) ERR("Unable to find group which is to be merged");
		
		original->merge(other);
		
		FORALL(other->getUnits(),uit) {
			unitGroupMap[*uit] = original;
			
			// ASSUMED IT WILL BE SET LATER, BUT WE WILL REINITIALIZE HERE JUST IN CASE
			ortsUnits[*uit] = ExtraObjectData();
		}
		
		other->getUnits().clear();

		Order *currentOrder = original->getOrder();
		
		if (currentOrder && !currentOrder->isCompleted())  currentOrder->executed = false;
		
		//groups.erase(git);
		//delete other;
	}	
	
	real4 Player::getInfluence(NUS_TerrainBase::Loc loc){
		if(useInfluence){
			return  attack_influence.value_at(loc.x, loc.y);
		}
		else
			return 0;
	}
	
	real8 Player::getVictoryValue()  {
		switch (victoryType) {
			case 0:
				if (bases.empty()) 
					return -1.0;
				else
					return (real8)bases.size();
			case 1:
				if (ortsUnits.empty())
					return -1.0;
				else
					return (real8)ortsUnits.size();
			default:
				break;
		}
		
		return 0.0;
	}
	
	void Player::removeOrtsUnit(Unit* obj) { 
		// Task::removeUnit will handle removing unit from group
		if (obj == NULL) ERR("Trying to remove a NULL object");
		
		if (currentStrategy) {
			currentStrategy->removeUnit(obj);
		}
		else {
			removeInternalUnit(obj);
		}

		if(useInfluence){
			attack_influence.add_circle(lastInfluencePos[obj].x , lastInfluencePos[obj].y, 112,-0.5);
			attack_influence.add_circle(lastInfluencePos[obj].x , lastInfluencePos[obj].y, 224,-0.5);
		}
		
		if (ortsUnits.erase(obj) == 0) ERR("removeOrtsUnit: unable to find obj for ortsUnits"); 

		if (objectUnitMap.erase(obj->getObject()) == 0) ERR("removeOrtsUnit: unable to find obj for objectUnitMap");

		if (unitGroupMap.erase(obj) == 0) ERR("removeOrtsUnit: unable to find obj for unitGroupMap");


		delete obj;		
	}
	
	void Player::addOrtsBase(GameObj *obj){
		Base *newBase = new Base(obj, playerID);
		
		ortsBases[newBase] = ExtraObjectData();
		objectBaseMap[obj] = newBase;
		bases.push_back(newBase);
		// we can add a base right away, since there is no grouping of bases, they are what they are
	}

	void Player::removeObject(GameObj *obj) {
		if (objectBaseMap.find(obj) != objectBaseMap.end())
			removeOrtsBase(obj);
		else
		{
			if (objectUnitMap.find(obj) != objectUnitMap.end())
				removeOrtsUnit(obj);
		}
	}
	
	void Player::removeOrtsBase(GameObj *obj) {
		Base *base = objectBaseMap[obj];
		
		removeOrtsBase(base);
	}
	
	void Player::removeOrtsBase(Base* obj){ 
		ortsBases.erase(obj); 
		objectBaseMap.erase(obj->getObject());
		
		Vector<Base*>::iterator it = std::find(bases.begin(), bases.end(), obj);
		
		if (it != bases.end()) bases.erase(it);
		
		delete obj;
	}

}
