#include <algorithm>

#include "NUS_Task.h"

#include "NUS_SimulationAI.H"

#include "NUS_SimulationState.h"

static const bool debug = false;

namespace NUS_Simul {

	MoveStruct::MoveStruct(const NUS_TerrainBase::Loc& p) {
		pos = p;
		base = NULL;
		unit = NULL;
	}

	MoveStruct::MoveStruct(Base *base) {
		pos = NUS_TerrainBase::Loc(base->getPosition().x,base->getPosition().y);
		this->base = base;
		unit = NULL;
	}
	
	MoveStruct::MoveStruct(Unit *unit) {
		pos = NUS_TerrainBase::Loc(unit->getPosition().x,unit->getPosition().y);
		this->unit = unit;
		base = NULL;
	}
	
	Task::Task()
	{
		parent = NULL;
		state = NULL;
		playerID = 0;
		lineNum = 0;
		function = NULL;
		extraData = NULL;

		startTime = getCurrentTime();
		timeout = 0.0;
	}

// this is the task constructor for when we have a parent

	Task::Task(Task* parent, GroupVector& groupVector,
		TaskCreateFunctionPtr function, std::string name,real8 timeout,
		MoveStruct* extraData)
	{
		if(parent == NULL) {
			ERR("we should always have a parent task in this Task constructor!!!");
		}
		this->parent = parent;
		playerID = parent->playerID;
		state = parent->state;

		if(function == NULL) {
			ERR("we should never have a NULL function in task constructor!");
		}
		this->function = function;
		this->name = name;
		this->timeout = timeout;
		this->extraData = extraData;

		lineNum = 0;

		if(groupVector.empty()) {
			ERR("warning,  trying to create a task with no objects, shouldn't be doing this!");
		}

		groups = groupVector;

		// add ourselves as a child of our parent
		
		// now we remove all the objects this child gets from the parent
		FORALL(groups,it) {
			//if (debug) std::cout <<  "child taking " << (*it) << std::endl;
			Vector<Group*>::iterator it2 = std::find(parent->groups.begin(), parent->groups.end(), *it);

			if (it2 != parent->groups.end()) 
			{
				parent->groups.erase(it2);
				
				//if (debug) std::cout <<  "remaining" << std::endl;
				//FORALL(parent->groups,it3) 
				//	if (debug) std::cout <<  "\t" << (*it3) << std::endl;
					
			}
			else	
				ERR("BAD, group in parent should always exist when giving it to a child");
		}
		
		parent->children.push_back(this);

		startTime = getCurrentTime();
	}

// when we don't have a parent, this is the task constructor
	Task::Task(uint4 playerID,SimulationState* state,GroupVector& origGroups, TaskCreateFunctionPtr function,
	std::string name,real8 timeout,MoveStruct* extraData) {

		this->playerID = playerID;

		if(state == NULL) {
			ERR("warning: trying to set a null state for task, this should not happen");
		}
		this->state = state;
		this->parent = NULL;
		children.clear();

		if(origGroups.empty()) {
			ERR("warning, trying to create a subtask with no groups, shouldn't be doing this!");
		}

		groups = origGroups;
		
		if(function == NULL) {
			ERR("we should never have a NULL function in parentless task constructor!");
		}
		
		this->function = function;
		this->name = name;

		this->timeout = timeout;
		this->extraData = extraData;
		lineNum = 0;

		startTime = getCurrentTime();
	}

	Task::Task(const Task& t) {
		groups = t.groups;
		playerID = t.playerID;
		name = t.name;
		startTime = t.startTime;
		timeout = t.timeout;
		function = t.function;
		lineNum = t.lineNum;
		if(t.extraData != NULL) {
			extraData = t.extraData->clone();
		}else{
			extraData = NULL;
		}
	}

	Task::~Task() {
		FORALL(children, it) {
			delete *it;
		}
		delete extraData;
		children.clear();

		// group deletion 
		// Option 1 involves player (see NUS_Player.h)
		// actually we should not be deleting any groups usually,
		// as they are copied out by childCompleted
		Player* currPlayer = state->getPlayer(playerID);

		FORALL (groups, it) 
			currPlayer->removeGroup(*it);

		groups.clear();		
	}

	void Task::addGroup(Group *group) {
		groups.push_back(group);
		
		state->getPlayer(playerID)->addGroup(group);
	}
	
	void Task::removeGroup(Group *group) {
		Vector<Group*>::iterator it = std::find(groups.begin(), groups.end(), group);
		
		if (it != groups.end()) {
			groups.erase(it);
			
			state->getPlayer(playerID)->removeGroup(group);
		}
		else {
			// maybe it's in a sub-task
			for (Vector<Task*>::iterator it = children.begin(); it != children.end(); it++)
				(*it)->removeGroup(group);
		}
	}
	
	void Task::removeAllGroups() {
		for (Vector<Group*>::iterator it = groups.begin(); it != groups.end(); it++) 
			state->getPlayer(playerID)->removeGroup(*it);

		groups.clear();
	}
	
	// DO NOT CALL THIS WITHIN TASKS
	// This is used by Player to remove a unit from a group
	bool Task::removeUnit(Unit *unit) {
		for (Vector<Group*>::iterator it = groups.begin(); it != groups.end(); it++) {
			Vector<Unit*> &groupUnits((*it)->getUnits());
			
			Vector<Unit*>::iterator uit = std::find(groupUnits.begin(), groupUnits.end(), unit);
			
			if (uit != groupUnits.end()) {								
				groupUnits.erase(uit);
				
				if (groupUnits.empty()) 
					removeGroup(*it);					
				
				return true;
			}
			
		}
		
		// maybe it's in a sub-task
		for (Vector<Task*>::iterator it = children.begin(); it != children.end(); it++) 
			if ((*it)->removeUnit(unit)) return true;
			
		return false;
	}

	void Task::childCompleted(Task* child) {
		assert(child != NULL);

	// add our child's objects back to the parent once the child is done
		FORALL(child->groups,it) {
			groups.push_back(*it);
		}

		child->groups.clear();
	}

	bool Task::update() {
		//if (debug) std::cout <<  "NOW UPDATING " << name << std::endl;
		
		bool childFinished = false;
		TaskVector::iterator it= children.begin();
		while(it != children.end()) {
			if((*it)->update()) {
				childCompleted(*it);
				delete *it;
				childFinished = true;
				it = children.erase(it);
			}else{
				++it;
			}
		}

	// if we have no children, and no groups, we are definitely done!
		if(groups.empty() && children.empty()) return true;

	// if we're out of time, then everyone is done, and mark them so...(not sure if marking is needed, but better
	// to be safe than sorry
		if(timeout > 0.0 && (getCurrentTime() - startTime) > timeout){
			
			if (debug) std::cout <<  "TIMEOUT FOR TASK " << name << std::endl;
			
			if (children.empty()) {
				Player* currPlayer = state->getPlayer(playerID);
			
				FORALL(groups,it) {
					//Vector<Group*> &playerGroups(currPlayer->getGroups());
				
					//Vector<Group*>::iterator git = std::find(playerGroups.begin(), playerGroups.end(), *it);
				
					//if (git != playerGroups.end()) {
					for (Vector<Unit*>::iterator uit = (*it)->getUnits().begin(); uit != (*it)->getUnits().end(); uit++)
						currPlayer->getOrtsUnits()[*uit].isDone = true;
					
					if ((*it)->getOrder()) (*it)->getOrder()->setCompleted(true);
					
					//if ((*it)->getOrder()) (*it)->cleanupOrder();
					//}				
				}
			
				return true;
			}
		}

	//if we have children, but no groups, then we are not done
		if(groups.empty() && !children.empty()) return false;
		
		if(!groups.empty() && childFinished  /*&& children.empty()*/) { 
			// only return here if task explicitly kills itself
			//fprintf(stderr,"sanka my ass\n");
			if(exec()){
				//if(!children.empty()) ERR("NOT IMPLEMENTED YET. need to properly kill all children as well when we do a hard abort");
				return true;
			}
		}
		
		if(!children.empty()) return false;

		//if (debug) std::cout <<  "Testing to see if all groups are done" << std::endl;

		FORALL(groups,it) {
			if ((*it)->getOrder() && (!(*it)->getOrder()->isCompleted())) return false;
		}
		
		// clean up all orders
		FORALL(groups,it) 
			if ((*it)->getOrder()) (*it)->cleanupOrder();
			
		if (debug) std::cout <<  "Done Testing to see if all groups are done" << std::endl;

		return true;
	}

	void Task::merge(Group *original, Group *other) {
		/*
		Vector<Group*>::iterator git = std::find(groups.begin(), groups.end(), other);
		
		/if (git == groups.end()) {
			//ERR("Unable to find group which is to be merged");
			state->getPlayer(playerID)->currentStrategy->removeGroup(other);
		}
		else
			groups.erase(git);
		*/
		
		state->getPlayer(playerID)->merge(original, other);
		state->getPlayer(playerID)->getStrategy()->removeGroup(other);
	}

}
