#ifndef Game1_ST_PFEngine_H
#define Game1_ST_PFEngine_H

// $Id: Game1_ST_PFEngine.H,v 1.1 2006/12/23 18:32:19 orts_furtak Exp $

// This is an ORTS file 
// (c) Michael Buro, Sami Wagiaalla, David Deutscher
// licensed under the GPL

#include "Game1_ST_SimpleMap.H"
#include "Game1_ST_Bucket.H"
#include "Game1_SimpleTerrain.H"

#include "boost/smart_ptr.hpp"

namespace Game1_SimpleTerrain {

  // ------------------------------------------------------------------------------
  /** Adapt SimpleMap to use a given Resolution Factor. 
      The map will represent each RF X RF integer coordinate points of 
      the world as a single integer coordinate.
      It is assumed that w_ and h_ are divisible by RF.
      All of SimpleMap's cell access functions assume input coords are given
      in the *inner* scale, and not in the world scale.

      The main point for this intermediate class is code clarity and a joint
      interface for AStar maps and SizeCell maps.
  */
  template <typename T>
  struct ArbitraryResolutionSimpleMap : public SimpleMap<T>
  {
    /** 
    Construct an empty map according to given size specifications.
    All T cells in the map are constructed using their default ctor.
    @param w_                 Width (in *world* coordinates) of the map.
    @param h_                 Height (in *world* coordinates) of the map.
    @param resolution_factor_ Factor of decreased resolution; 1 => no decrease.
    */
    ArbitraryResolutionSimpleMap(sint4 w_, sint4 h_, sint4 resolution_factor_)
        // Keep a down-sized internal map, (+1 for what? find out/ask Sami)
      : SimpleMap<T>(w_/resolution_factor_ + 1, h_/resolution_factor_ + 1),
        resolution_factor(resolution_factor_), disp(resolution_factor_/2)
    {
      assert( (w_ % resolution_factor)==0 && "Width must be divisible by resolution_factor");
      assert( (h_ % resolution_factor)==0 && "Height must be divisible by resolution_factor");
    };

    //---------------------------------------------------------------------
    /// \name Coordinate translation: world <-> inner representation
    //---------------------------------------------------------------------
    //@{
    sint4 world2x(sint4 worldx) const { return (worldx+disp) / resolution_factor; };
    sint4 world2y(sint4 worldy) const { return (worldy+disp) / resolution_factor; };
    sint4 x2world(sint4 x) const { return x * resolution_factor; };
    sint4 y2world(sint4 y) const { return y * resolution_factor; };
    sint4 world2r(sint4 worldr) const {
      sint4 r = worldr / resolution_factor;
      return (worldr%resolution_factor >= (resolution_factor+1)/2)? r+1 : r;
    };
    /// This is max[ abs( c - x2world(world2x(c)) )]
    sint4 max_error_from_conversion() const { return resolution_factor/2; };
    //@}
    //---------------------------------------------------------------------

  protected:
    // For coordinate translations
    sint4 resolution_factor;      ///< how many world points per one point in our representation
    sint4 disp;                   ///< objects are displaced half an (internal) cell so boundaries fall in the middle of tiles
  };


// ------------------------------------------------------------------------------
// ------------------------------------------------------------------------------
// ------------------------------------------------------------------------------

  class ST_Terrain::PFEngine {

  // ------------------------------------------------------------------------------
    /** Extend the SimpleMap with set/access methods to allow multiple levels-of-detail.
        This factors out much of the old functionality of PFEngine, and the 
        implementation is tightly coupled to PFEngine, so it's an internal class.
    */
    class Map : public ArbitraryResolutionSimpleMap<SizeCell> {

    public:
      Map(sint4 w_, sint4 h_, sint4 resolution_factor_) 
        : ArbitraryResolutionSimpleMap<SizeCell>(w_, h_, resolution_factor_) {};

      //---------------------------------------------------------------------
      /// \name Changes to the world.
      //---------------------------------------------------------------------
      //@{
      void insert_object(const Object *obj);
      void remove_object(const Object *obj);
      void update_object(const Object *obj);
      void insert_boundary(const Segment& line);

      // Insert/remove the given path (on the ground) as an obstacle to avoid.
      // Expects a forward order of waypoints, as returned by TerrainBase::find_path().
      void ir_path(const Object *obj, const Path &path, bool insert, bool with_endpoints);
      //@}
      //---------------------------------------------------------------------

    private:
      //---------------------------------------------------------------------
      /// \name Internal machinery for updating the map
      //---------------------------------------------------------------------
      //@{
      void ir_object(const Object *obj, bool insert, 
        bool use_prev_attributes=false);                              ///< Expects world coordinates
      void insert_rectangle(const Loc &topleft,const Loc &bottomright, 
                            sint4 cost);                              ///< Expects world coordinates

      void insert_bucket(const Bucket &b, sint4 delta_maxsize, 
                         sint4 cost);                                 ///< Expects inner coordinates
      void insert_line(sint4 x1, sint4 y1, sint4 x2, sint4 y2,
                       sint4 delta_maxsize, sint4 cost);              ///< Expects inner coordinates
      void set_cell(sint4 x, sint4 y,
                    sint4 delta_maxsize, sint4 cost);                 ///< Expects inner coordinates

      /** false iff the cost for the given object to stay at the given
          location will not make A* report failure. Works only for CIRCLEs. */
      //bool is_blocked(const Object *obj, const Loc &world_location);  ///< Expects world coordinates
      //@}
      //---------------------------------------------------------------------

      // prevent copies
      Map(const Map &);
      Map& operator=(const Map &);
    };

    friend class Map;

// ------------------------------------------------------------------------------
// ------------------------------------------------------------------------------

  public:
    typedef uint4 LevelOfDetail;  ///< index into the LOD array passed to the ctor

    /**
    @param tiles_x_       Width of world in tiles.
    @param tiles_y_       Height of world in tiles.
    @param tile_points_   How many points per tile are in the full world representation.
    @param lods_needed    Vector indicating levels of detail of inner world representation needed.
                          1 - full details, resolution as in the world rep;
                          2 - half resolution; 4 - quarter resultion; etc.
                          (tiles_x_*tile_points_) and (tiles_y_*tile_points_) must be divisible by each LOD.
    */
    PFEngine(sint4 tiles_x_, sint4 tiles_y_, sint4 tile_points_, const Vector<sint4>& lods_needed);

    //---------------------------------------------------------------------
    /// \name Changes to the world.
    //---------------------------------------------------------------------
    //@{
    /** Insert a given object to the internal world representation.
        Pointers must remain valid until a respective remove_object() call. */
    void insert_object(const Object *obj);
    void remove_object(const Object *obj);
    void update_object(const Object *obj);
    void insert_boundary(const Segment& line);

    // Insert/remove the given path (on the ground) as an obstacle to avoid.
    // Expects a forward order of waypoints, as returned by TerrainBase::find_path().
    void ir_path(const Object *obj, const Path &path, bool insert, bool with_endpoints);
    //@}
    //---------------------------------------------------------------------


    //---------------------------------------------------------------------
    /// \name Perform path finding 
    //---------------------------------------------------------------------
    //@{
    /** Perform pathfinding to a location.
        @param obj        Object to move, starting with it's current position.
        @param goal       Target location.
        @param output     If found a path: this will hold it; guaranteed to 
                          include at least one waypoint for the goal. 
                          Waypoints are listed in reverse order, goal is in locs[0]. 
        @param distance   Consider points that are at this distance from 'goal' as targets.
                          (must be >= 0 to get a successful result).
        @param lod        Work with the given resolution in the world representation, 
                          allows accuracy-time tradeoffs.
        @param consider   Ignore some or all mobile objects as obstacles for the path.
        @param stop_time  Stop and report failure if and when SDL_GetTicks() > stop_time.

        @return True iff path found, and the path itself in 'output'. */
    bool find_path(const Object *obj, const Loc &goal, Path& output, sint4 distance = 0, LevelOfDetail lod = 0, ConsiderObjects consider = CONSIDER_ALL, uint4 stop_time = 0);
    /** Perform pathfinding to a target object. */
    bool find_path(const Object *obj, const Object *goal, Path& output, sint4 distance = 0, LevelOfDetail lod = 0, ConsiderObjects consider = CONSIDER_ALL, uint4 stop_time = 0);
    /** Perform pathfinding on a small submap, to an object. */
    bool sub_path(const Object *obj, const Loc &topLeft, const Loc &bottomRight, const Object *goal_obj, Path &output, sint4 distance = 0, LevelOfDetail lod = 0, ConsiderObjects consider = CONSIDER_ALL, uint4 stop_time = 0);
    /** Perform pathfinding on a small submap, to a location. */
    bool sub_path(const Object *obj, const Loc &topLeft, const Loc &bottomRight, const Loc &goal_loc, Path &output, sint4 distance = 0, LevelOfDetail lod = 0, ConsiderObjects consider = CONSIDER_ALL, uint4 stop_time = 0);
    //@}
    //---------------------------------------------------------------------

    // temporary hack - we want to plan a path from the center of the CC (in case some
    // of its sides are blocked). For this we will order "ignore_rectangels()". 
    // A variation might be considered for the TerrainBase interface after the tournament.
    // These must be called in pairs, and consider_* must be restored before the next
    // view is processed.
    void ignore_rectangles(void);
    void consider_rectangles(void);

  private:

    typedef ArbitraryResolutionSimpleMap<AStarCell> AStarMap;
    typedef Vector< boost::shared_ptr<Map> >        MapsVector;

    Vector<sint4> lods;                       ///< Levels Of Detail available.
    LevelOfDetail max_details_LOD;            ///< The most detailed level.
    MapsVector    ground_maps;
    MapsVector    air_maps;
    Vector< boost::shared_ptr<AStarMap> >  astar_maps;
    std::set<const Object*> mobile_objects;  ///< Set of all mobile objects, allows ignoring them during PFing
    std::set<const Object*> rectangle_objects;  ///< Set of all rectangular objects, allows ignoring them during PFing


    /// Initialize a given map to be empty with a blocked border.
    static void init_sizes(SimpleMap<SizeCell> &smap);

    //---------------------------------------------------------------------
    /// \name Actual computation of paths
    //---------------------------------------------------------------------
    //@{
    /// Various params for controlling the execution of simple_astar()
    struct ControlParams {
      uint4 stop_time;  ///< stop and report failure when SDL_GetTicks() >= stop_time; 0 => no time limit
      uint4 stop_cost;  ///< stop and report failure when expanding a node with cost higher than this; 0 => no cost limit

      // default values - no time limit; stop when expanding an occupied node.
      ControlParams(uint4 stop_time_ = 0, 
                    uint4 stop_cost_ = (uint4)((UNIT_COST < BOUND_COST)? UNIT_COST:BOUND_COST))
        : stop_time(stop_time_), stop_cost(stop_cost_) {};
    };

    /** This implementation can handle multiple target locations and can 
        handle a target object. To this end it accepts a predicate for 
        testing whether a specific map location (inner coords + map pointer) 
        is an allowable goal location.
        It currently doesn't support a pluggable distance huristic, but 
        instead accepts 'the goal' location (probably the center of the 
        target) and uses the usual octile distance huristic. 
        This combination results in an inadmissible huristic, but should 
        give reasonable results without resorting to the slower huristic
        h = min distance to all goal locations. */
    template <class GoalTester>
    bool find_path(const Object *obj,             ///< the object to move
                   const Loc &start_position,     ///< Move it from where (not necessarily its current position)
                   const Loc &goal,               ///< the goal location used for distance estimation
                   const GoalTester &goal_tester,        ///< a predicate for an AStarCell which is true iff it is a goal location
                   const Loc &topleft,            ///< topleft corner of the submap to explore for this path; use Loc(0,0) to specify the entire map
                   const Loc &bottomright,        ///< bottomright corner of the submap to explore for this path; use Loc(-1,-1) or smaller to specify the entire map
                   LevelOfDetail lod,             ///< the LOD to use
                   ConsiderObjects consider,      ///< what objects to consider/ignore as obstacles
                   Path& output,                  ///< output path is returned here
		   const ControlParams &control_params);

    /** \brief The actual A* implementation.
        Return true iff a path was found, and the goal location (cell index in astar_map)
        that was flagged by 'goal_tester' in 'end_of_path'. 
        Only exapnd nodes in the submap defined by 'topleft' and 'bottomright'. */
    template<class GoalTester>
    bool simple_astar(const Map &world_view, 
                      AStarMap &astar_data,
                      Bucket to_move, 
                      sint4 x2, sint4 y2,
                      const GoalTester &goal_tester,
                      const Loc &topleft,
                      const Loc &bottomright,
                      uint4 &end_of_path,
		      const ControlParams &control_params);

    /// Remove objects we wish to ignore for PFing from the world map
    void ignore_objects(ConsiderObjects consider, boost::shared_ptr<Map> map, const Object *moved_obj);
    /// Return the objects we ignored for the last PFing back to the world map
    void reconsider_objects(ConsiderObjects consider, boost::shared_ptr<Map> map, const Object *moved_obj);

    /// Remove redundant waypoints (that fall on a straight line)
    void smoothen_path(Vector<Loc> &path);
    /// Compute the octile distance between two points
    uint4 estimate(sint4 x1, sint4 y1, sint4 x2, sint4 y2);
    /// Check whether the given three points are on a straight line
    bool on_line(const Loc &a, const Loc &b, const Loc &c);
    //@}
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    /// \name Predicates for testing arrival to different types of goals
    //---------------------------------------------------------------------
    //@{
    /** Helper predicate for testing whether moving an object to a given location
        will place it adjacent to a given target, considering the shapes and sizes 
        of the object and target. */
    class IsNearObject; friend class IsNearObject;

    /** Helper predicate for testing whether a tested location is equal to the 
        target location. */
    class IsNearLocation; friend class IsNearLocation;

    /** Helper predicate for going to the closest clear location, i.e. one
        with no other overlapping objects (might happen when using a low resolution). */
    class IsAtClearLocation; friend class IsAtClearLocation;
    //@}
    //---------------------------------------------------------------------

    //---------------------------------------------------------------------
    /// \name Cost constants
    //---------------------------------------------------------------------
    //@{
    static const uint4 FACTOR = 128; ///< distance multiplier (real->int)
    static const uint4 DIAG   = sint4(FACTOR*1.4142);
    /// cost of passing though a square occupied by another unit
    static const sint4 UNIT_COST = 1000000;   // signed to allow decreasing the cost of a cell
    /// cost of passing though a square occupied by and orts boundery
    static const sint4 BOUND_COST = 5000000;  // signed to allow decreasing the cost of a cell
    /** cost of passing though a square within the direction of
    * another units x-second distance
    * x-second distance: if a unit continues to move in a straight
    * line it will be there in x-seconds */
    //static const uint4 CROSS_COST = 500;
    //@}
    //---------------------------------------------------------------------

    /// Serving the graphical widget that is used for developement.
    boost::shared_ptr<const Map> get_ground_map() const { return ground_maps[0]; };
    boost::shared_ptr<const Map> get_air_map() const { return air_maps[0]; };

    // prevent copies
    PFEngine(const PFEngine &other);
    PFEngine& operator=(const PFEngine &other);

    friend class SimpleTerrainWidget;
    friend class SimpleTerrainOverlay;
  };

} // end of namespace SimpleTerrain

#endif
