/** @file Collision2D.C
    @see Collision2D.H

    @todo Speed improvement (pythagoras) possible!

    $Id: Collision2D.C,v 1.7 2006/06/13 17:25:40 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Collision2D.C,v $
*/

// This is an ORTS file
// (c) Tim Furtak    
// (c) Michael Buro  
// licensed under the GPL

#include "Collision2D.H"

using namespace std;
using namespace collision2d;

#define COLL_PAR 0   // print collision parameters

template <class T> T sgn(const T& x)
{
  if (x >= 0) return 1;
  return -1;
}

real8 collision2d::no_collision = 1e6;
real8 collision2d::now = 0.0;

//===================================================================

/** Solves \f$ at^2 + bt = c; a > 0 \f$
    @return true iff solution exists

    robust solution (numerical recipies)
*/

bool collision2d::solve_quadratic(real8 a, real8 b, real8 c, real8 &t1, real8 &t2)
{
#if COLL_PAR
  sint4 u = 4096*4096;
  cout << "solve: " << a/u << " " << b/u << " " << c/u << endl;
#endif
  
  if (a < 0) ERR("a < 0");
  if (a == 0) {
    if (b == 0) return false;
    t1 = t2 = -c / b;
    return true;
  }

  c = -c;
  real8 d = b*b-4*a*c;
  if (d < 0) return false;

  // robust root computation (Numerical Recipies p.184)
  
  real8 q = -0.5L * (b + sgn(b) * sqrt(d));
  t1 = q / a;
  t2 = c / q;

  if (t1 > t2) swap(t1, t2);
  assert(!(t1 > t2));
  
#ifndef NDEBUG
  real8 e1 = fabs(a*t1*t1+b*t1+c);
  real8 e2 = fabs(a*t2*t2+b*t2+c);
  if (e1 > 1.5e-5) cout << "error: " << e1 << " " << t1 << " " << a << endl;
  if (e2 > 1.5e-5) cout << "error: " << e2 << " " << t2 << " " << a << endl;
#endif
  
  return true;
}

//===================================================================

/** Collision test: moving circle (c,r,h) and line end point (lx,ly).
    @return true iff collision occurs in time interval [u,v].
    rr = r*r
    @param a       length of circle's heading squared
    @param center  location of the circle
    @param h       circle's heading
    @param rr      circle's radius squared
    @param u       start of time interval
    @param v       end of time interval
    @param lx      point x coord
    @param ly      point y coord
    @param t       set to collision time
*/

bool collision2d::collision_outside(real8 a,
                                    const ScalarPoint &center,
                                    const ScalarPoint &h,
                                    real8 rr, real8 u, real8 v,
                                    const Scalar &lx, const Scalar &ly,
                                    real8 &t)
{
  // solve a *t^2 + b* t <= c
  // a = 1/inva = (hx*hx+hy*hy)
  // b = (hx*(cx-lx)+hy(cy-ly)
  // c = r*r - (cx-lx)**2 - (cy-ly)**2

#if COLL_PAR
  cout << "out " << a << " " << center.x << " " << center.y << " "
       << h.x << " " << h.y << " "
       << rr << " " << u << " " << v << " " << lx << " " << ly << endl;
#endif
  
  real8 b = 2.0 * (mult(sint8(h.x), sint8(center.x-lx)) +
                   mult(sint8(h.y), sint8(center.y-ly)));
  real8 c = rr - square(sint8(center.x-lx)) - square(sint8(center.y-ly));
  real8 t1, t2;

  if (!solve_quadratic(a, b, c, t1, t2)) return false;

  if (t1 == t2) return false;

  // penetration in [t1,t2]
  assert(!(t1 > t2));
  // intersect with [u,v]
  if (t1 > v || t2 < u) return false;  // empty intersection
  if (t1 <= u) {                       // u in [t1,t2]
    //  ---t1---u---min-------t2-----
    if (u == 0 && u >= (t1+t2)*0.5) return false; // moving away
    t = u; return true;
  }
  t = t1; return true;                 // t1 > u -> t1
}

//===================================================================

/** Collision test: moving circle (c,r,h) and line x=lx.
    @return true iff collision occurs in time interval [u,v].
*/

bool collision2d::collision_inside_x(const ScalarPoint &c, const ScalarPoint &h,
                                     const Scalar &r, real8 u, real8 v,
                                     const Scalar &lx, real8 &t)
{
  assert(r > 0);

  if (h.x == 0) {
    if (abs(c.x - lx) <= r) { t = u; return true; }
    return false;
  }
  
  // abs(c.x + t * h.x - x) <= r

  real8 invhx = 1.0/h.x;
  real8 t1 = ( r-c.x+lx)*invhx;
  real8 t2 = (-r-c.x+lx)*invhx;
  if (t1 > t2) swap(t1,t2);

  // intersect [t1,t2] with [u,v]

  if (t1 > v || t2 < u) return false; // no intersection
  if (t1 <= u) { // u in [t1,t2]
    t = u; return true;
  }
  t = t1; return true;  // t1 > u -> t1
}

//===================================================================

/** Collision test: moving circle (c,r,h) and line y=ly.
    @return true iff collision occurs in time interval [u,v].
*/

bool collision2d::collision_inside_y(const ScalarPoint &c, const ScalarPoint &h,
                               const Scalar &r, real8 u, real8 v,
                               const Scalar &ly, real8 &t)
{
  assert(r > 0);

  if (h.y == 0) {
    if (abs(c.y - ly) <= r) { t = u; return true; }
    return false;
  }
  
  // abs(c.y + t * h.y - y) <= r

  real8 invhy = 1.0/h.y;
  real8 t1 = ( r-c.y+ly)*invhy;
  real8 t2 = (-r-c.y+ly)*invhy;
  if (t1 > t2) swap(t1,t2);
  
  // intersect [t1,t2] with [u,v]

  if (t1 > v || t2 < u) return false; // no intersection
  if (t1 <= u) { // u in [t1,t2]
    t = u; return true;
  }
  t = t1; return true;  // t1 > u -> t1
}

//===================================================================

/** Compute collision time of a (moving) circle and a (moving) point.
    Used to check for collision with the endpoint of a line.
    @param c     circle center
    @param rr    square of the circle's radius
    @param p     point's location
    @param h     circle's heading relative to the point
    @param a     length of the heading squared
    @param time  start simulation at this time
*/

real8 collision2d::collision_time_circle_point(const ScalarPoint &c, real8 rr,
                                               const ScalarPoint &p,
                                               const ScalarPoint &h, real8 a,
                                               real8 time)
{
  ScalarPoint c_p(c.x-p.x, c.y-p.y);

  real8 qc = rr - square(c_p.x) - square(c_p.y);

  Scalar c_h = c_p * h; // c dot h
  if (c_h > 0) {   // moving away from each other
    if (qc >= 0) { // they don't intersect
      return no_collision;
    }
  }

  real8 t1, t2;

  real8 qa = a; // square(h.x) + square(h.y);
  real8 qb = 2 * c_h;

  bool coll = solve_quadratic(qa, qb, qc, t1, t2); // (a)t^2 + (b)t = c

  if (!coll) return no_collision;
  if (t1 == t2) return no_collision; // point intersects tangentially

  if (t2 <= time) return no_collision;
  if (t1 <= time) return now; // intersecting at time

  return t1;
}

//===================================================================

/** Compute collision time of a (moving) circle and a (moving) point.
    Used to check for collision with the endpoint of a line.
    @param c     position of circle
    @param rr    square of the circle's radius
    @param h1    circle's heading
    @param p     position of point
    @param h2    point's heading
    @param time  start simulation at this time
*/

real8 collision2d::collision_time_circle_point(const ScalarPoint &c, real8 rr,
                                               const ScalarPoint &h1,
                                               const ScalarPoint &p,
                                               const ScalarPoint &h2,
                                               real8 time)
{
  ScalarPoint h(h1.x-h2.x, h1.y-h2.y);
  ScalarPoint c_p(c.x-p.x, c.y-p.y);

  real8 qc = rr - square(c_p.x) - square(c_p.y);

  Scalar c_h = c_p * h; // c dot h
  if (c_h > 0) {   // moving away from each other
    if (qc >= 0) { // they don't intersect
      return no_collision;
    }
  }

  real8 t1, t2;

  real8 qa = square(h.x) + square(h.y);
  real8 qb = 2 * c_h;

  bool coll = solve_quadratic(qa, qb, qc, t1, t2); // (a)t^2 + (b)t = c

  if (!coll) return no_collision;
  if (t1 == t2) return no_collision; // point intersects tangentially

  if (t2 <= time) return no_collision;
  if (t1 <= time) return now; // intersecting at time

  return t1;
}

//===================================================================

/** Compute collision time of a moving circle and a still hvline.
    Moves circle to time ct first and scales heading difference accordingly.
    @return Collision time relative to ct.
*/

real8 collision2d::collision_time_circle_hvline(const ScalarPoint &c, real8 r_,
                                                const ScalarPoint &h1,
                                                const ScalarPoint &a1,
                                                const ScalarPoint &a2,
                                                const ScalarPoint &h2,
                                                sint4 &coll_type, real8 time)
{
  ScalarPoint h(h1.x-h2.x, h1.y-h2.y);
  coll_type = -1; // uncomputed
  real8 rr = r_*r_;
  Scalar r = Scalar(r_);

  if (a1.x == a2.x) {

    // vline |

    Scalar lx = a1.x; // line x

    // let the object drift away from the line if it got stuck there (the equality check might be too much)
    if (c.x <= lx && h.x <= 0) return no_collision;
    if (c.x >= lx && h.x >= 0) return no_collision;

    Scalar ly1 = a1.y;
    Scalar ly2 = a2.y;

    real8 a = square(sint8(h.x)) + square(sint8(h.y));
    real8 t;

    if (h.y != 0) {
      
      if (h.y < 0) swap(ly1, ly2);

      real8 hyinv = 1.0/h.y;
      real8 t1 = real8(ly1-c.y)*hyinv; // time to cross y1
      real8 t2 = real8(ly2-c.y)*hyinv; // time to cross y2
      
      // [-oo, t1] ^ [time,1]
      if (t1 >= time) {
        real8 t = collision_time_circle_point(c, rr, ScalarPoint(lx, ly1), h, a, time);
        if (t != no_collision) {
          coll_type = (h.y < 0) ? 3 : 1; // first point
          return t;
        }
      }

      // [t1,t2] ^ [ct,1]
      if (t2 >= time) {
        real8 u = t1, v = t2;
        if (u < time) u = time;
        if (collision_inside_x(c, h, r, u, v, lx, t)) {
          coll_type = 2;
          return t;
        }
      }

      // [t2,oo] ^ [ct,1]
      {
        real8 t = collision_time_circle_point(c, rr, ScalarPoint(lx, ly2), h, a, time);
        if (t != no_collision) {
          coll_type = (h.y < 0) ? 1 : 3; // second point
          return t;
        }
      }

      return no_collision;

    } else {

      // h.y == 0

      if (c.y < ly1) {
        coll_type = 1;
        return collision_time_circle_point(c, rr, ScalarPoint(lx, ly1), h, a, time);
      } else if (c.y <= ly2) {
        if (collision_inside_x(c, h, r, time, 1.0, lx, t)) {
          coll_type = 2;
          return t;
        }
      } else {
        coll_type = 3;
        return collision_time_circle_point(c, rr, ScalarPoint(lx, ly2), h, a, time);
      }
      return no_collision;
    }

  } else {

    // hline --

    Scalar ly = a1.y;

    if (c.y <= ly && h.y <= 0) return no_collision;
    if (c.y >= ly && h.y >= 0) return no_collision;

    Scalar lx1 = a1.x;
    Scalar lx2 = a2.x;

    real8 a = square(sint8(h.x)) + square(sint8(h.y));
    real8 t;

    if (h.x != 0) {

      if (h.x < 0) swap(lx1, lx2);

      real8 hxinv = 1.0/h.x;
      real8 t1 = real8(lx1-c.x)*hxinv; // time to cross x1
      real8 t2 = real8(lx2-c.x)*hxinv; // time to cross x2

      // [-oo, t1] ^ [ct,1]
      if (t1 >= time) {
        real8 t = collision_time_circle_point(c, rr, ScalarPoint(lx1, ly), h, a, time);
        if (t != no_collision) {
          coll_type = (h.x < 0) ? 3 : 1; // first point
          return t;
        }
      }

      // [t1,t2] ^ [ct,1]
      if (t2 >= time) {
        real8 u = t1, v = t2;
        if (u < time) u = time;
        if (collision_inside_y(c, h, r, u, v, ly, t)) {
          coll_type = 2;
          return t;
        }
      }

      // [t2,oo] ^ [ct,1]
      {
        real8 t = collision_time_circle_point(c, rr, ScalarPoint(lx2, ly), h, a, time);
        if (t != no_collision) {
          coll_type = (h.x < 0) ? 1 : 3; // second point
          return t;
        }
      }

      return no_collision;

    } else {

      // h.x == 0

      if (c.x < lx1) {
        coll_type = 1;
        return collision_time_circle_point(c, rr, ScalarPoint(lx1, ly), h, a, time);
      } else if (c.x <= lx2) {
        if (collision_inside_y(c, h, r, time, 1.0, ly, t)) {
          coll_type = 2;
          return t;
        }
      } else {
        coll_type = 3;
        return collision_time_circle_point(c, rr, ScalarPoint(lx2, ly), h, a, time);
      }
      return no_collision;
    }
  }

  ERR("got here?");
  return 0;
}

//===================================================================

real8 collision2d::collision_time_circle_line(const ScalarPoint &c, real8 r,
                                              const ScalarPoint &h1,
                                              const ScalarPoint &a1, const ScalarPoint &a2,
                                              const ScalarPoint &h2,
                                              sint4 &coll_type, real8 time)
{
  Scalar x0 = a1.x;
  Scalar x1 = a2.x;
  Scalar y0 = a1.y;
  Scalar y1 = a2.y;

  if (x0 == x1 || y0 == y1)
    return collision_time_circle_hvline(c, r, h1, a1, a2, h2, coll_type, time);

  ScalarPoint lp0 = a1;
  //  ScalarPoint  lp1(x1,y1);

  ScalarPoint h(h1.x-h2.x, h1.y-h2.y); // relative heading

  ScalarPoint l(x1-x0, y1-y0);         // vector from point 1 to point 2
  ScalarPoint l2c(c.x - x0, c.y - y0); // vector from point 1 to circle

  // cross product -> distance from the center to the infinite line
  real8 dist2line = real8(l.cross(l2c)) / l.l2_norm();
  real8 d2l = (dist2line < 0) ? -dist2line : dist2line; // abs value

  // check if it's not touching the line and it's moving away
  if (d2l > r) { // used to be > 0

    // (signed) distance of the heading to the line (only need the sign, skip division)
    sint8 h_dist = l.cross(h);
    if (dist2line * h_dist >= 0) return no_collision; // not moving closer to the line
    //    if (dist2line < 0 && h_dist <= 0) return no_collision;
    //    if (dist2line > 0 && h_dist >= 0) return no_collision;
  }

  sint4 end1_marker = 1; // first endpoint
  sint4 end2_marker = 3; // second endpoint

  real8 alpha = real8(h*l) / real8(l*l); // parallel speed as a fraction of the line vector
  if (alpha < 0) { // heading and line are pointing in opposite directions
    alpha = -alpha;
    l.scale(-1);
    swap(x0,x1);
    swap(y0,y1);
    swap(end1_marker,end2_marker);
  }

  real8 speed = h.l2_norm();
  real8 vdist = alpha * l.l2_norm(); // = speed * beta;
  //real8 beta = vdist / speed;        // fraction of motion parallel to the line
  real8 zdist = sqrt(speed*speed - vdist*vdist);


  real8 rr = r*r;
  real8 a = speed*speed;
  real8 t;

  // parallel distance from the first point to the object, measured as a multiple of the line vector
  real8 u = (static_cast<sint8>(c.x-x0)*(x1-x0) + static_cast<sint8>(c.y-y0)*(y1-y0)) /
    static_cast<real8>(l.l2_norm_sq());

  if (alpha == 0) { // find the section, do the check

    if (u <= 0) {
      if (collision_outside(a, c, h, rr, time, 1.0, x0, y0, t)) {
        coll_type = end1_marker;
        return t;
      }
      return no_collision;
    }

    if (u >= 1) {
      if (collision_outside(a, c, h, rr, time, 1.0, x1, y1, t)) {
        coll_type = end2_marker;
        return t;
      }
      return no_collision;
    }

    t = (d2l-r) / zdist; // time to hit the line
    if (t < 0) t = 0.0;
    if (t > 1) t = no_collision;
    coll_type = 2;
    return t;
  }

  real8 t1,t2,t3; // cross first endpoint, line collision, cross last endpoint

  if (u <= 0) { // the closest point is before the interval (check the endpoint)
    t1 = (-u)*l.l2_norm() / speed; // time until the center crosses into the mid-range
    t1 += time;
    real8 v = t1;
    if (v > 1.0) v = 1.0;
    if (collision_outside(a, c, h, rr, time, v, x0, y0, t)) {
      coll_type = end1_marker;
      return t;
    }

    if (t1 >= 1.0) return no_collision; // doesn't reach the mid-range
  }


  t3 = ((1-u)*l.l2_norm()) / vdist; // amount of line left / parallel speed

  if (t3 < 0) { // already past the mid-section
    t2 = 2;
    t3 = time;
  } else {
    t2 = (d2l-r) / zdist; // time to hit the line
    if (t2 < 0) t2 = 0;
    if (t2 > 1) t2 = 2;
  }

  if (t2 <= 1.0) {
    if (t2 <= t3) { // check if it hits the line before passing the endpoint
      coll_type = 2;
      return t2;
    }
  }


  if (t3 <= 1.0) {
    if (collision_outside(a, c, h, rr, time, 1.0, x1, y1, t)) {
      coll_type = end2_marker;
      return t;
    }
  }

  return no_collision;

}

//===================================================================

real8 collision2d::collision_time_circle_rect(const ScalarPoint &c, real8 r,
                                              const ScalarPoint &h1,
                                              const ScalarPoint &a1,
                                              const ScalarPoint &a2,
                                              const ScalarPoint &h2,
                                              sint4 &coll_type, real8 time)
{
  ScalarPoint h(h1.x-h2.x, h1.y-h2.y);
  if (c.x+r < a1.x && h.x <= 0) return 2.0;
  if (c.x-r > a2.x && h.x >= 0) return 2.0;
  if (c.y+r < a1.y && h.y <= 0) return 2.0;
  if (c.y-r > a2.y && h.y >= 0) return 2.0;


  real8 time1, time2;
  sint4 coll_type1 = 0, coll_type2 = 0; // vertical line, horizontal line

  // vertical
  
  if (c.x < a1.x) {
    ScalarPoint b(a1.x, a2.y);
    time1 = collision_time_circle_line(c, r, h1, a1, b, h2, coll_type1, time);
    coll_type1 += 10;
  }
  else if (c.x > a2.x) {
    ScalarPoint b(a2.x, a1.y);
    time1 = collision_time_circle_line(c, r, h1, b, a2, h2, coll_type1, time);
    coll_type1 += 20;
  } else {
    time1 = collision2d::no_collision;
  }

  // horizontal

  if (c.y < a1.y) {
    ScalarPoint b(a2.x, a1.y);
    time2 = collision_time_circle_line(c, r, h1, a1, b, h2, coll_type2, time);
    coll_type2 += 30;
  }
  else if (c.y > a2.y) {
    ScalarPoint b(a1.x, a2.y);
    time2 = collision_time_circle_line(c, r, h1, b, a2, h2, coll_type2, time);
    coll_type2 += 40;
  } else {
    time2 = collision2d::no_collision;
  }

  if (time1 < time2) {
    coll_type = coll_type1;
    return time1;
  } else {
    coll_type = coll_type2;
    return time2;
  }
}

//===================================================================

/* Computes the intersection point of two lines: the line object, and the
   moving point. */

real8 collision2d::collision_time_line_point(const ScalarPoint &a1,
                                             const ScalarPoint &a2,
                                             const ScalarPoint &h1,
                                             const ScalarPoint &p,
                                             const ScalarPoint &h2,
                                             sint4 &/*coll_type*/, real8 time)
{
  ScalarPoint h(h2.x-h1.x, h2.y-h1.y); // relative heading

  const Scalar x1 = a1.x;
  const Scalar x2 = a2.x;
  const Scalar y1 = a1.y;
  const Scalar y2 = a2.y;

  if (h.x == 0 && h.y == 0) return no_collision; // no relative position change

  const Scalar x3 = p.x;
  const Scalar y3 = p.y;

  real8 inv_norm = 1.0 / (square(x2-x1) + square(y2-y1));

  real8 dist_y = (x2-x1)*(y3-y1) - (y2-y1)*(x3-x1); // perp distance from the point to the line
  real8 dist_x = (x2-x1)*(x3-x1) + (y2-y1)*(y3-y1); // projected x position of the point along the line
  dist_x *= inv_norm;

  real8 dy = (x2-x1)*h.y - (y2-y1)*h.x; // rate the point moves perp. to the line
  real8 dx = (x2-x1)*h.x + (y2-y1)*h.y; // rate the point moves parallel to the line
  dx *= inv_norm;

  if (abs(dx) < 1e-8) { // point is moving (almost) perpendicular to the line

    if (dist_x < 0 || dist_x > 1) return no_collision;
    if (dist_y == 0) return 0.0;
    if (abs(dy) < 1e-8) return no_collision;
    real8 ty = -dist_y / dy; // time to cross the infinite line
    if (ty < time) return no_collision;
    return ty;

  } else {

    if (dy == 0) {
      if (dist_y == 0) {
        if (dist_x >= 0 && dist_x <= 1) return 0.0;
        real8 t1 = -dist_x / dx;      // time to cross the first point
        real8 t2 = -(dist_x-1.0) / dx; // time to cross the second point
        if (t2 < t1) swap(t1, t2);
        if (t1 < time) {
          if (t2 < time) return no_collision;
          return t2;
        }
        return t1;
      }
      return no_collision;
    }

    real8 t1 = -dist_x / dx;      // time to cross the first point
    real8 t2 = -(dist_x-1.0) / dx; // time to cross the second point

    if (t2 < t1) swap(t1, t2);

    real8 ty = -dist_y / dy; // time to cross the infinite line

    if (ty < t1 || ty > t2) return no_collision;
    if (ty < time) return no_collision;
    return ty;
  }
}

//===================================================================

real8 collision2d::collision_time_line_line(const ScalarPoint &a1,
                                            const ScalarPoint &a2,
                                            const ScalarPoint &h1,
                                            const ScalarPoint &b1,
                                            const ScalarPoint &b2,
                                            const ScalarPoint &h2,
                                            sint4 &coll_type, real8 time)
{
  ScalarPoint h(h1.x-h2.x, h1.y-h2.y); // relative heading

  coll_type = -1; // not used yet

  ScalarPoint line1  (a1.x - a2.x, a1.y - a2.y);
  ScalarPoint line1_1(a1.x - b1.x, a1.y - b1.y); // line to the first point of line 2
  ScalarPoint line1_2(a1.x - b2.x, a1.y - b2.y); // line to the second point of line 2

  ScalarPoint line2  (b1.x - b2.x, b1.y - b2.y);
  ScalarPoint line2_1(b1.x - a1.x, b1.y - a1.y); // line to the first point of line 1
  ScalarPoint line2_2(b1.x - a2.x, b1.y - a2.y); // line to the second point of line 1

  // compute if the lines cross each other's extensions

  real8 side_1_1 = line1.cross(line1_1);
  real8 side_1_2 = line1.cross(line1_2);

  real8 side_2_1 = line2.cross(line2_1);
  real8 side_2_2 = line2.cross(line2_2);

  bool cross1, cross2;

  // fixme: equality condition is somewhat arbitrary, might want to change it
  cross1 = (side_1_1 * side_1_2 <= 0); // line 2 crosses line 1's extension
  cross2 = (side_2_1 * side_2_2 <= 0); // line 1 crosses line 2's extension

  if (cross1 && cross2) return 0.0; // the lines intersect

  // centroid of line 2
  ScalarPoint c2((b1.x + b2.x) / 2,
           (b1.y + b2.y) / 2);

  ScalarPoint line_c(c2.x - a1.x, c2.y - a1.y); // start of line 1 to the centroid of line 2
  real8 side_1_c = line1.cross(line_c);
  real8 side_1_h = line1.cross(h);

  if (side_1_c * side_1_h < 0) return no_collision; // the centroid is moving away from the line

  // Since the lines don't intersect, we only need to check if the endpoints
  // cross into another line within the current step.

  real8 t1, t2, t3, t4;
  sint4 temp;

  t1 = collision_time_line_point(b1, b2, h2, a1, h1, temp, time);
  t2 = collision_time_line_point(b1, b2, h2, a2, h1, temp, time);
  t3 = collision_time_line_point(a1, a2, h1, b1, h2, temp, time);
  t4 = collision_time_line_point(a1, a2, h1, b2, h2, temp, time);

  return min( min(t1, t2), min(t3, t4) );
}

//===================================================================

real8 collision2d::collision_time_line_rect(const ScalarPoint &a1,
                                            const ScalarPoint &a2,
                                            const ScalarPoint &h1,
                                            const ScalarPoint &b1,
                                            const ScalarPoint &b2,
                                            const ScalarPoint &h2,
                                            sint4 &coll_type, real8 time)
{
  ScalarPoint h(h1.x-h2.x, h1.y-h2.y); // relative heading (line relative to the rectangle)

  real8 mt = collision2d::no_collision;

  ScalarPoint r1(b1.x, b1.y);
  ScalarPoint r2(b1.x, b2.y);
  ScalarPoint r3(b2.x, b2.y);
  ScalarPoint r4(b2.x, b1.y);

  real8 t;
  sint4 temp;
  if (h.x < 0) {
    t = ::collision_time_line_line(a1, a2, h1, r4, r3, h2, temp, time);
    if (t < mt) { mt = t; coll_type = 10 + temp; }
  }
  if (h.x > 0) {
    t = ::collision_time_line_line(a1, a2, h1, r1, r2, h2, temp, time);
    if (t < mt) { mt = t; coll_type = 20 + temp; }
  }
  if (h.y < 0) {
    t = ::collision_time_line_line(a1, a2, h1, r2, r3, h2, temp, time);
    if (t < mt) { mt = t; coll_type = 30 + temp; }
  }
  if (h.y > 0) {
    t = ::collision_time_line_line(a1, a2, h1, r1, r4, h2, temp, time);
    if (t < mt) { mt = t; coll_type = 40 + temp; }
  }

  return mt;
}

//===================================================================

real8 collision2d::collision_time_rect_rect(const ScalarPoint &a1,
                                            const ScalarPoint &a2,
                                            const ScalarPoint &h1,
                                            const ScalarPoint &b1,
                                            const ScalarPoint &b2,
                                            const ScalarPoint &h2,
                                            sint4 &coll_type, real8 time)
{
  ScalarPoint h(h1.x-h2.x, h1.y-h2.y); // relative heading (1st relative to the 2nd)

  bool int_x = false;
  bool int_y = false;

  if (a1.x <= b1.x) {
    if (a2.x > b1.x) int_x = true;
  } else {
    if (b2.x > a1.x) int_x = true;
  }

  if (a1.y <= b1.y) {
    if (a2.y > b1.y) int_y = true;
  } else {
    if (b2.y > a1.y) int_y = true;
  }

  if (int_x && int_y) return collision2d::now;

  //------------------------------

  real8 mt = collision2d::no_collision;

  ScalarPoint ar1(a1.x, a1.y);
  ScalarPoint ar2(a1.x, a2.y);
  ScalarPoint ar3(a2.x, a2.y);
  ScalarPoint ar4(a2.x, a1.y);

  ScalarPoint br1(b1.x, b1.y);
  ScalarPoint br2(b1.x, b2.y);
  ScalarPoint br3(b2.x, b2.y);
  ScalarPoint br4(b2.x, b1.y);

  real8 t;
  sint4 temp;
  if (h.x < 0) {
    t = ::collision_time_line_line(ar1, ar2, h1, br4, br3, h2, temp, time);
    if (t < mt) { mt = t; coll_type = 10 + temp; }
  }
  if (h.x > 0) {
    t = ::collision_time_line_line(ar4, ar3, h1, br1, br2, h2, temp, time);
    if (t < mt) { mt = t; coll_type = 20 + temp; }
  }
  if (h.y < 0) {
    t = ::collision_time_line_line(ar1, ar4, h1, br2, br3, h2, temp, time);
    if (t < mt) { mt = t; coll_type = 30 + temp; }
  }
  if (h.y > 0) {
    t = ::collision_time_line_line(ar2, ar3, h1, br1, br4, h2, temp, time);
    if (t < mt) { mt = t; coll_type = 40 + temp; }
  }

  return mt;
}

//===================================================================
//===================================================================

sint4 collision2d::resolve_collision_circle_point(const ScalarPoint &c, real8 /*rr*/,
                                                  const ScalarPoint &h1,
                                                  const ScalarPoint &p, const
                                                  ScalarPoint &h2,
                                                  real8 time)
{
  sint4 stopped = 0;

  ScalarPoint c1(Scalar(c.x + h1.x * time), Scalar(c.y + h1.y * time));
  ScalarPoint c2(Scalar(p.x + h2.x * time), Scalar(p.y + h2.y * time));

  c2.sub(c1); // collision normal

  sint8 dot1 = c2 * h1; // need to use sint8 or similar here
  sint8 dot2 = c2 * h2;

  if (dot1 >= 0) stopped += 1;
  if (dot2 <= 0) stopped += 2;

  return stopped;
}

//===================================================================

sint4 collision2d::resolve_collision_circle_circle(const ScalarPoint &c1, real8 /*rr1*/,
                                                   const ScalarPoint &h1,
                                                   const ScalarPoint &c2, real8 /*rr2*/,
                                                   const ScalarPoint &h2,
                                                   real8 time)
{
  return resolve_collision_circle_point(c1, 0, h1, c2, h2, time);
}

//===================================================================

sint4 collision2d::resolve_collision_circle_line(const ScalarPoint &c, real8 r,
                                                 const ScalarPoint &h1,
                                                 const ScalarPoint &a1,
                                                 const ScalarPoint &a2,
                                                 const ScalarPoint &h2,
                                                 sint4 coll_type, real8 time)
{
  sint4 stopped = 0;

  if (coll_type == 1) { // first endpoint

    stopped = resolve_collision_circle_point(c, r*r, h1, a1, h2, time);

  } else if (coll_type == 3) { // second endpoint

    stopped = resolve_collision_circle_point(c, r*r, h1, a2, h2, time);

  } else if (coll_type == 2) { // line proper

    ScalarPoint c1 (Scalar(c.x + h1.x * time),
              Scalar(c.y + h1.y * time)); // circle position at time
    ScalarPoint l11(Scalar(a1.x + h2.x * time),
              Scalar(a1.y + h2.y * time)); // start of the line at time
    ScalarPoint lc (c1.x - l11.x, c1.y - l11.y); // circle pos relative to line endpoint at time

    ScalarPoint l(a2.x - a1.x, a2.y - a1.y); // the line vector

    sint8 cross_side = l.cross(lc); // the side of the line the collision occurred on
    sint8 cross_c_head = l.cross(h1); // which side of the line the circle is heading towards
    sint8 cross_l_head = l.cross(h2); // which side of the line the line is heading towards

    if (cross_side * cross_c_head < 0) stopped += 1; // circle is moving towards the line
    if (cross_side * cross_l_head > 0) stopped += 2; // line is moving towards the circle

  } else {
    cout << "unknown circle-line collision type " << coll_type << endl; 
  }

  return stopped;
}

//===================================================================

sint4 collision2d::resolve_collision_circle_rect(const ScalarPoint &c, real8 r,
                                                 const ScalarPoint &h1,
                                                 const ScalarPoint &a1,
                                                 const ScalarPoint &a2,
                                                 const ScalarPoint &h2,
                                                 sint4 coll_type, real8 time)
{
  sint4 coll_seg = coll_type / 10;
  coll_type %= 10;

  // left, right, bottom, top

  switch (coll_seg) {
  case 1: return resolve_collision_circle_line(c, r, h1, a1, ScalarPoint(a1.x, a2.y),
                                               h2, coll_type, time);

  case 2: return resolve_collision_circle_line(c, r, h1, ScalarPoint(a2.x, a1.y), a2,
                                               h2, coll_type, time);
    
  case 3: return resolve_collision_circle_line(c, r, h1, a1, ScalarPoint(a2.x, a1.y),
                                               h2, coll_type, time);
    
  case 4: return resolve_collision_circle_line(c, r, h1, ScalarPoint(a1.x, a2.y), a2,
                                               h2, coll_type, time);
  default:
    ERR("got here?");
  }
  return 0;
}

//===================================================================

