/** @file Motion.C
    @see Game.H

    $Id: Motion.C,v 1.40 2007/04/02 23:24:52 marcar Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/kernel/src/Motion.C,v $
*/

// This is an ORTS file
// (c) Michael Buro
// (c) Tim Furtak
// licensed under the GPL

/*
  Collision categories:

  (ON_SURFACE = ON_LAND || ON_WATER)

  UNDER_WATER: collision with WATER/LAND boundary + FRINGE
  ON_WATER:    collision with ON_WATER objects and WATER/LAND boundary + FRINGE
  ON_LAND:     collision with ON_LAND objects and WATER/LAND and GROUND +
               FRINGE boundaries
  IN_AIR:      collisions with FRINGE (and IN_AIR objects)


  simplified:

  UNDER_WATER: collision with all UNDER_WATER objects and all boundaries
  ON_WATER:    collision with all ON_WATER objects and all boundaries
  ON_LAND:     collision with all ON_LAND objects and all boundaries
  IN_AIR:      collisions with FRINGE (and IN_AIR objects)

*/

#include "Game.H"
#include "GameObj.H"
#include "ServerObjData.H"
#include "GUI.H"
#include "PairHeap.H"
#include "CycleCounter.H"
#include "Array.H"
#include "ServerObjData.H"
#include "GameTile.H"
#include "Options.H"
#include "Profiler.H"

using namespace std;


#define TEST_OUT 0
#define TEST_OBJ 0
#define ZCAT_CHECK 0

#define TOO_CLOSE(x, a, b)                                              \
      {                                                                 \
        cout << "TOO CLOSE " << x << " " << a << " " << b << endl;      \
 /* fixme: no direct gui access anymore  if (gui) gui->display(a, b); FOREVER{}; exit(1);    */            \
      }

//===================================================================

void MotionSectors::init(sint4 num_sec)
{
  assert(num_sec > 0);
  num = num_sec;

  MotionSectorInfo si;
  FORS (i, num) {
    secs.push_back(si);
    sizes.push_back(0);
  }

  if (active) delete [] active;
  active = new uint1[num];
}

//-------------------------------------------------------------------

void MotionSectors::reset_sizes() {
  // cout << secs.size() << " " << sizes.size() << endl;
  assert(secs.size() == sizes.size());
  FORU (i, secs.size()) secs[i].set_size(sizes[i]);
}

//-------------------------------------------------------------------

void MotionSectors::clear() {
  secs.clear(); sizes.clear();
}

//-------------------------------------------------------------------

void MotionSectors::insert(GameObj *g, sint4 s, bool is_static)
{
  assert(s >= 0 && s < num);
  if (!is_static && !active[s]) {
    active[s] = true;
    active_sectors.push_back(s);
  }
  
  MotionSectorInfo &ms = secs[s];

  // The motion sectors could have non-boundary objects in them.
  // The boundary needs to be inserted (anywhere) before those, with the other static objects.

  if (is_static) {
    ms.push_back(g);
    sint4 i = sizes[s]; // original number of boundary objects
    swap(ms[i], ms[ms.size()-1]);
    ++sizes[s];
  } else {
    ms.push_back(g); // fixme: check if duplicates will cause a problem
  }
}

//===================================================================

/** Compute connected component by following neighborhood pointers.
    Store pointers to game objects in pointer array pointed to by comp.
    The resulting network can be checked for moving objects by
    Game::move_objects(Array<GameObj*> &mobjs, MotionSectors &ms, bool obj_coll).
*/

void Game::visit(GameObj &e, GameObj** &comp)
{
#if TEST_OBJ
  if (!e.test_magic_ok()) {
    ERR("oops");
  }
#endif

  if (e.smd.visit_magic == visit_magic) return; // already visited

  e.smd.visit_magic = visit_magic;
  *comp++ = &e;

  const Neighbors &ne = e.smd.neighbors;

  FORS (ni, ne.size()) visit(*ne[ni], comp);
}

//===================================================================

/** Create a boundary object, add it to the relevant movement layers.
    Can be called after motion sector initialization has already taken place;
    places the boundary with the other static objects at the beginning of
    the list and updates the sector size.

    @todo Compute the motion sectors a boundary is added to with a line drawing
          algorithm instead of a rectangle.  Same for add_tile_boundary, and
          remove_tile_boundary.

    @param x  x position of line endpoint 1, given in tile units.
    @param y  y position of line endpoint 1, given in tile units.
    @param x2 x position of line endpoint 2, given in tile units.
    @param y2 y position of line endpoint 2, given in tile units.

    @param is_fringe Whether the boundary is part of the game field hard limits.
*/

GameObj* Game::add_tile_boundary(sint4 x, sint4 y, sint4 x2, sint4 y2,
                                 Tile::Type t1, Tile::Type t2, bool height_diff, bool is_fringe)
{
  // not really needed, but it keeps the line's bounding circle small
  assert(x-x2 <= 1 && x-x2 >= -1);
  assert(y-y2 <= 1 && y-y2 >= -1);

  ScriptObj *temp = NPC->gof.create("boundary");
  GameObj *gob = temp->get_GameObj();
  assert(gob);
  if (!gob) ERR("out of memory");
  gob->init_sod();

  sint4 xp  = x  * tile_points;
  sint4 yp  = y  * tile_points;
  sint4 x2p = x2 * tile_points;
  sint4 y2p = y2 * tile_points;

  *gob->sod.x1 = xp;
  *gob->sod.y1 = yp;
  *gob->sod.x2 = x2p;
  *gob->sod.y2 = y2p;
  gob->sod.is_boundary = true;

  *gob->sod.shape = Object::LINE;

  if (x == x2 || y == y2) {
    sint4 rad = ::max(x2p - xp, y2p - yp);
    gob->smd.bc_fine_r = fine_grid(rad)/2;
  } else {
    gob->smd.bc_fine_r = Scalar(fine_grid(tile_points) * 0.71) +1;
  }

  gob->smd.bc_fine_center.x = fine_grid(*gob->sod.x1 + *gob->sod.x2)/2;
  gob->smd.bc_fine_center.y = fine_grid(*gob->sod.y1 + *gob->sod.y2)/2;

  *gob->sod.ncollides = 0;
  if (is_fringe) {
    *gob->sod.collides = (uint4)(-1);
    *gob->sod.zcat = Object::ZCAT_UNDEF;
  } else {
    if (height_diff) {
      *gob->sod.collides = tile_height_diff_mask;
    } else {
      *gob->sod.collides = tile_collision_mask[num_tile_types * t1 + t2];
    }
    *gob->sod.zcat = Object::ON_LAND;
  }

  boundaries.push_back(gob);

  // compute sectors

  sint4 sx1 = fine_grid(xp)/sec_wh;
  sint4 sy1 = fine_grid(yp)/sec_wh;
  sint4 sx2 = fine_grid(x2p)/sec_wh;
  sint4 sy2 = fine_grid(y2p)/sec_wh;

  // take the rectangle of motion sectors and mark them
  // a line drawing algorithm would be better

  if (sx2 < sx1) swap(sx1,sx2);
  if (sy2 < sy1) swap(sy1,sy2);

  if (sx1 < 0) {
    sx1 = 0;
    if (sx2 < 0) sx2 = 0;
  }
  if (sx2 >= sec_nx) {
    sx2 = sec_nx-1;
    if (sx1 >= sec_nx) sx1 = sec_nx-1;
  }

  if (sy1 < 0) {
    sy1 = 0;
    if (sy2 < 0) sy2 = 0;
  }
  if (sy2 >= sec_ny) {
    sy2 = sec_ny-1;
    if (sy1 >= sec_ny) sy1 = sec_ny-1;
  }

  for (sint4 i = sx1; i <= sx2; ++i) {
    for (sint4 j = sy1; j <= sy2; ++j) {
      sint4 ind = j * sec_nx + i;

      if (is_fringe) {
        fringe_secs.secs[ind].push_back(gob);
      }

      boundary_secs.secs[ind].push_back(gob);
      boundary_secs.sizes[ind]++;

      motion_secs.insert(gob, ind, true);
    }
  }

  return gob;
}

//===================================================================

/** Remove a boundary aligned on the tile grid.
    Maintains the validity of the motion sectors affected by the removal.

    @param x  x position of line endpoint 1, given in tile units.
    @param y  y position of line endpoint 1, given in tile units.
    @param x2 x position of line endpoint 2, given in tile units.
    @param y2 y position of line endpoint 2, given in tile units.
*/

void Game::remove_tile_boundary(sint4 x, sint4 y, sint4 x2, sint4 y2)
{
  sint4 xp  = x  * tile_points;
  sint4 yp  = y  * tile_points;
  sint4 x2p = x2 * tile_points;
  sint4 y2p = y2 * tile_points;

  // compute sectors

  sint4 sx1 = fine_grid(xp)/sec_wh;
  sint4 sy1 = fine_grid(yp)/sec_wh;
  sint4 sx2 = fine_grid(x2p)/sec_wh;
  sint4 sy2 = fine_grid(y2p)/sec_wh;

  // take the rectangle of motion sectors and mark them
  // a line drawing algorithm would be better

  if (sx2 < sx1) swap(sx1,sx2);
  if (sy2 < sy1) swap(sy1,sy2);

  if (sx1 < 0) sx1 = 0; else if (sx2 >= sec_nx) sx2 = sec_nx-1;
  if (sy1 < 0) sy1 = 0; else if (sy2 >= sec_ny) sy2 = sec_ny-1;

  for (sint4 i = sx1; i <= sx2; ++i) {
    for (sint4 j = sy1; j <= sy2; ++j) {
      sint4 ind = j * sec_nx + i;
      sint4 ind2, ind3;

      FORS(i, boundary_secs.sizes[ind]) {
        MotionSectorInfo &msi = boundary_secs.secs[ind];
        ServerObjData &b = msi[i]->sod;
        if (*b.x == xp &&
            *b.y == yp &&
            *b.x2 == x2p &&
            *b.y2 == y2p) {
          ind2 = boundary_secs.sizes[ind]-1;  // last boundary in the list
          ind3 = boundary_secs.secs.size()-1; // last object in the list
          msi[i] = msi[ind2];
          msi[ind2] = msi[ind3];
          msi.set_size(ind3);
          boundary_secs.sizes[ind]--;
          break;
        }
      }

      FORS(i, motion_secs.sizes[ind]) {
        MotionSectorInfo &msi = motion_secs.secs[ind];
        ServerObjData &b = msi[i]->sod;
        if (*b.x == xp &&
            *b.y == yp &&
            *b.x2 == x2p &&
            *b.y2 == y2p) {
          ind2 = motion_secs.sizes[ind]-1;  // last boundary in the list
          ind3 = motion_secs.secs.size()-1; // last object in the list
          msi[i] = msi[ind2];
          msi[ind2] = msi[ind3];
          msi.set_size(ind3);
          motion_secs.sizes[ind]--;
          break;
        }
      }
    }
  }
  
}


//===================================================================

/** Check the surrounding tiles and add/remove boundaries as necessary. */

void Game::update_tile(sint4 i, Vector<ScriptObj::WPtr> *new_boundaries)
{
  sint4 x = pf.ind2x(i);
  sint4 y = pf.ind2y(i);

  GameTile &t = pf(i);

  if (t.get_split() == Tile::TB_SPLIT && !(t.boundaries & (1 << 4))) {
    if (t.boundaries & (1 << 5)) {
      remove_tile_boundary(x,y+1,x+1,y);
      t.boundary_int = 0;
    }
    if ((t.get_typeW() != t.get_typeE()) || t.is_half()) {
      bool height_diff = t.is_half();
      GameObj *b = add_tile_boundary(x, y, x+1, y+1, t.get_typeW(), t.get_typeE(), height_diff);
      t.boundaries |= (1 << 4);
      t.boundary_int = b;
      if (new_boundaries) new_boundaries->push_back(b);
    }
  }
  else if (t.get_split() == Tile::BT_SPLIT && !(t.boundaries & (1 << 5))) {
    if (t.boundaries & (1 << 4)) {
      remove_tile_boundary(x,y,x+1,y+1);
      t.boundary_int = 0;
    }
    if ((t.get_typeW() != t.get_typeE()) || t.is_half()) {
      bool height_diff = t.is_half();
      GameObj *b = add_tile_boundary(x, y+1, x+1, y, t.get_typeW(), t.get_typeE(), height_diff);
      t.boundaries |= (1 << 5);
      t.boundary_int = b;
      if (new_boundaries) new_boundaries->push_back(b);
    }
  }


  if (x > 0) {
    GameTile &t1 = pf(x-1,y);
    GameTile &t2 = pf(x,y);

    bool height_diff = (t1.get_height(2) != t2.get_height(7) ||
                        t1.get_height(3) != t2.get_height(6));

    if (t1.get_typeE() != Tile::UNKNOWN && t2.get_typeW() != Tile::UNKNOWN &&
        (height_diff || t1.get_typeE() != t2.get_typeW())) {
      
      if (!(t1.boundaries & (1 << 1))) {
        GameObj *b = add_tile_boundary(x, y, x, y+1, t1.get_typeE(), t2.get_typeW(), height_diff);
        t1.boundaries |= (1 << 1);
        t2.boundaries |= (1 << 3);
        t1.boundary_E = t2.boundary_W = b;
        if (new_boundaries) new_boundaries->push_back(b);
      }
    } else {
      if (t1.boundaries & (1 << 1)) {
        remove_tile_boundary(x,y,x,y+1);
        t1.boundaries &= ~(1 << 1);
        t2.boundaries &= ~(1 << 3);
        t1.boundary_E = t2.boundary_W = 0;
      }
    }
  }
  if (x < pf.get_width()-1) {
    GameTile &t1 = pf(x,y);
    GameTile &t2 = pf(x+1,y);

    bool height_diff = (t1.get_height(2) != t2.get_height(7) ||
                        t1.get_height(3) != t2.get_height(6));

    if (t1.get_typeE() != Tile::UNKNOWN && t2.get_typeW() != Tile::UNKNOWN &&
        (height_diff || t1.get_typeE() != t2.get_typeW())) {

      if (!(t1.boundaries & (1 << 1))) {
        GameObj *b = add_tile_boundary(x+1, y, x+1, y+1, t1.get_typeE(), t2.get_typeW(), height_diff);
        t1.boundaries |= (1 << 1);
        t2.boundaries |= (1 << 3);
        t1.boundary_E = t2.boundary_W = b;
        if (new_boundaries) new_boundaries->push_back(b);
      }
    } else {
      if (t1.boundaries & (1 << 1)) {
        remove_tile_boundary(x+1,y,x+1,y+1);
        t1.boundaries &= ~(1 << 1);
        t2.boundaries &= ~(1 << 3);
        t1.boundary_E = t2.boundary_W = 0;
      }
    }
  }


  if (y > 0) {
    GameTile &t1 = pf(x,y-1);
    GameTile &t2 = pf(x,y);

    uint1 ty1, ty2;
    if (t1.get_split() == Tile::BT_SPLIT) ty1 = t1.get_typeE();
    else ty1 = t1.get_typeW();
    if (t2.get_split() == Tile::TB_SPLIT) ty2 = t2.get_typeE();
    else ty2 = t2.get_typeW();
    
    bool height_diff = (t1.get_height(4) != t2.get_height(1) ||
                        t1.get_height(5) != t2.get_height(0));

    if (ty1 != Tile::UNKNOWN && ty2 != Tile::UNKNOWN &&
        (height_diff || ty1 != ty2)) {

      if (!(t1.boundaries & (1 << 2))) {
        GameObj *b = add_tile_boundary(x, y, x+1, y, ty1, ty2, height_diff);
        t1.boundaries |= (1 << 2);
        t2.boundaries |= (1 << 0);
        t1.boundary_S = t2.boundary_N = b;
        if (new_boundaries) new_boundaries->push_back(b);
      }
    } else {
      if (t1.boundaries & (1 << 2)) {
        remove_tile_boundary(x,y,x+1,y);
        t1.boundaries &= ~(1 << 2);
        t2.boundaries &= ~(1 << 0);
        t1.boundary_S = t2.boundary_N = 0;
      }
    }
  }

  if (y < pf.get_height()-1) {
    
    GameTile &t1 = pf(x,y);
    GameTile &t2 = pf(x,y+1);

    uint1 ty1, ty2;
    if (t1.get_split() == Tile::BT_SPLIT) ty1 = t1.get_typeE();
    else ty1 = t1.get_typeW();
    if (t2.get_split() == Tile::TB_SPLIT) ty2 = t2.get_typeE();
    else ty2 = t2.get_typeW();
    
    bool height_diff = (t1.get_height(4) != t2.get_height(1) ||
                        t1.get_height(5) != t2.get_height(0));

    if (ty1 != Tile::UNKNOWN && ty2 != Tile::UNKNOWN &&
        (height_diff || ty1 != ty2)) {
      
      if (!(t1.boundaries & (1 << 2))) {
        GameObj *b = add_tile_boundary(x, y+1, x+1, y+1, ty1, ty2, height_diff);
        t1.boundaries |= (1 << 2);
        t2.boundaries |= (1 << 0);
        t1.boundary_S = t2.boundary_N = b;
        if (new_boundaries) new_boundaries->push_back(b);
      }
    } else {
      if (t1.boundaries & (1 << 2)) {
        remove_tile_boundary(x,y+1,x+1,y+1);
        t1.boundaries &= ~(1 << 2);
        t2.boundaries &= ~(1 << 0);
        t1.boundary_S = t2.boundary_N = 0;
      }
    }
  }
}

//===================================================================

/** Compute all boundary line segments and put pointers
    into fringe and boundary sectors.

    fringe: frame around playing field

    boundary: fringe + all line segments that separate tiles of
              different types

+-+
| ...
+

*/

void Game::generate_boundaries()
{
  //  init_motion_sectors();

  bool hd = false; // height difference

  if (rotated) { // angled fringe
    FORS(i, gtiles_x) {
      pf(i, gtiles_y+i).boundary_int = add_tile_boundary(i, gtiles_y+i, i+1, gtiles_y+i+1, 0, 0, hd, true);
      pf(gtiles_y+i, i).boundary_int = add_tile_boundary(gtiles_y+i, i, gtiles_y+i+1, i+1, 0, 0, hd, true);
    }
    FORS(i, gtiles_y) {
      pf(i, gtiles_y-i-1).boundary_int = add_tile_boundary(i, gtiles_y-i, i+1, gtiles_y-i-1, 0, 0, hd, true);
      pf(gtiles_x+i, pf.get_height()-i-1).boundary_int =
        add_tile_boundary(gtiles_x+i, pf.get_height()-i, gtiles_x+i+1, pf.get_height()-i-1, 0, 0, hd, true);
    }
  } else { // normal fringe
    FORS (x, pf.get_width()) {
      pf(x,0).boundary_N = add_tile_boundary(x, 0, x+1, 0,  0,0, hd, true);
      pf(x,pf.get_height()-1).boundary_S = add_tile_boundary(x,pf.get_height(),x+1,pf.get_height(), 0, 0, hd, true);
    }
    FORS (y, pf.get_height()) {
      pf(0,y).boundary_W = add_tile_boundary(0, y, 0, y+1,  0,0, hd, true);
      pf(pf.get_width()-1, y).boundary_E = add_tile_boundary(pf.get_width(),y,pf.get_width(),y+1, 0, 0, hd, true);
    }
  }

  FORS (y, pf.get_height()) { // internal tile boundaries
    FORS (x, pf.get_width()) {
      GameTile &t = pf(x,y);

      hd = t.is_half();
      
      if ((t.get_split() == Tile::TB_SPLIT) && ((t.get_typeW() != t.get_typeE())) || hd) {
        t.boundary_int = add_tile_boundary(x, y, x+1, y+1, t.get_typeW(), t.get_typeE(), hd, false);
        t.boundaries |= (1 << 4);  //fixme: 4?

      } else if ((t.get_typeW() != t.get_typeE()) || hd) {
        t.boundary_int = add_tile_boundary(x, y+1, x+1, y, t.get_typeW(), t.get_typeE(), hd, false);
        t.boundaries |= (1 << 5);  //fixme: 5? 
      }
    }
  }

  FORS (y, pf.get_height()) { // vertical walls
    FORS (x, pf.get_width()-1) {
      GameTile &t1 = pf(x,y);
      GameTile &t2 = pf(x+1,y);

      hd = (t1.get_height(2) != t2.get_height(7) ||
            t1.get_height(3) != t2.get_height(6));

      if (t1.get_typeE() != Tile::UNKNOWN && t2.get_typeW() != Tile::UNKNOWN &&
          (hd || t1.get_typeE() != t2.get_typeW())) {
        
        GameObj *b = add_tile_boundary(x+1, y, x+1, y+1, t1.get_typeE(), t2.get_typeW(), hd, false);
        t1.boundaries |= (1 << 1); // fixme: 1?
        t2.boundaries |= (1 << 3); // fixme: 3?
        t1.boundary_E = t2.boundary_W = b;
      }
    }
  }

  FORS (x, pf.get_width()) { // horizontal walls
    FORS (y, pf.get_height()-1) {
      GameTile &t1 = pf(x,y);
      GameTile &t2 = pf(x,y+1);

      uint1 ty1, ty2;
      if (t1.get_split() == Tile::BT_SPLIT) ty1 = t1.get_typeE(); else ty1 = t1.get_typeW();
      if (t2.get_split() == Tile::TB_SPLIT) ty2 = t2.get_typeE(); else ty2 = t2.get_typeW();

      hd = (t1.get_height(4) != t2.get_height(1) ||
            t1.get_height(5) != t2.get_height(0));

      if (ty1 != Tile::UNKNOWN && ty2 != Tile::UNKNOWN &&
          (hd || ty1 != ty2)) {

        GameObj *b = add_tile_boundary(x, y+1, x+1, y+1, ty1, ty2, hd, false);
        t1.boundaries |= (1 << 2);
        t2.boundaries |= (1 << 0);
        t1.boundary_S = t2.boundary_N = b;
      }
    }
  }

  FORS (i, sec_num) {
    fringe_secs.sizes[i]   = fringe_secs.secs[i].size();
    boundary_secs.sizes[i] = boundary_secs.secs[i].size();
    motion_secs.sizes[i]   = motion_secs.secs[i].size();
  }
}

//===================================================================

/** Put the object on the tile grid point closest to its fine position. */

void Game::obj_move_to_grid(GameObj &e) const
{
  ScalarPoint g_pos = e.smd.fine_pos;
  g_pos.round(fine_grid_res_bits);
  *e.sod.x = g_pos.x >> fine_grid_res_bits;
  *e.sod.y = g_pos.y >> fine_grid_res_bits;
}

//===================================================================

void Game::obj_stop(GameObj &e) const
{
  if (*e.sod.is_moving) {
    obj_move_to_grid(e);
    *e.sod.speed = 0;
    *e.sod.is_moving = 0;
    e.smd.fine_heading.x = 0;
    e.smd.fine_heading.y = 0;
  }
}

//===================================================================

// advance position fine_pos by t*heading_left
// set pos to round(fine_pos)

void Game::obj_move_and_stop(GameObj &e, real8 t) const
{
  obj_move(e, t);
  obj_stop(e);
}

//===================================================================

// advance position fine_pos by t*heading_left, adjust heading_left

void Game::obj_move(GameObj &e, real8 t) const
{
  real8 tsf = real8(e.smd.ticks_so_far) + t;

  real8 s = e.smd.inv_ttt * tsf; // fraction of distance travelled to the target

  if (s >= 1.0) {
    e.smd.fine_pos = e.smd.target_f_pos;
  } else {
    e.smd.fine_pos.x = e.smd.start_f_pos.x + Scalar(s*e.smd.f_delta.x);
    e.smd.fine_pos.y = e.smd.start_f_pos.y + Scalar(s*e.smd.f_delta.y);
  }

#if TEST_OUT
  cout << &e << " " << e.sod.is_boundary << " MOVED BY "; delta.write(); cout << endl;
  real8 h = square(sint8(delta.x))+square(sint8(delta.y));
  h = sqrt(h);
  h /= fine_grid_res;
  cout << "step= " << h << " " << *e.sod.speed << " " << t << endl;

  assert(e.smd.fine_pos.x >= 0 && e.smd.fine_pos.x < fine_grid(points_x));
  assert(e.smd.fine_pos.y >= 0 && e.smd.fine_pos.y < fine_grid(points_y));
#endif
}

//===================================================================

/** Set new target location and speed.
    @return true on success
*/

bool Game::obj_move_to(GameObj &e, const ScalarPoint &tpos, sint4 speed) const
{
  sintptr *im = e.get_int_ptr("is_mobile");
  if (!im || *im < 1) {
    obj_stop(e);
    return false;
  }

  if (tpos.bad()) {
    tpos.write(cerr);
    cerr << "target out of range" << endl;
    return false;
  }

  if (speed > *e.sod.max_speed) speed = *e.sod.max_speed;
  if (speed <= 0) {
    obj_stop(e);
    return false;
  }

  if (*e.sod.is_moving) obj_move_to_grid(e); // if the object is already moving

  e.smd.fine_pos.x = fine_grid(*e.sod.x);
  e.smd.fine_pos.y = fine_grid(*e.sod.y);

  e.smd.start_f_pos = e.smd.fine_pos;

  e.smd.target_f_pos.x = fine_grid(tpos.x);
  e.smd.target_f_pos.y = fine_grid(tpos.y);

  ScalarPoint delta(e.smd.target_f_pos.x - e.smd.start_f_pos.x,
              e.smd.target_f_pos.y - e.smd.start_f_pos.y); // vector to the target

  Scalar dist = delta.l2_norm();
  if (dist == 0) {
    obj_stop(e);
    return false;
  }

  e.smd.f_delta = delta;


  real8 inv_ttt = real8(fine_grid(speed))/dist; // 1 / time to reach target

  delta.scale(inv_ttt); // distance per tick
  if (delta.x == 0 && delta.y == 0) return false;

  e.smd.fine_heading = delta;
  e.smd.inv_ttt = inv_ttt;
  e.smd.ticks_so_far = 0;
  e.smd.ticks_total = sint4(ceil(1.0/inv_ttt));

  if (e.smd.ticks_total <= 0) ERR("ticks <= 0");


  *e.sod.speed = speed;
  *e.sod.is_moving = 1;

  // compute approximate heading:  1...HEADING_N (0 = stopped)
  sint4 heading = GameConst::encode_dir(delta.x, delta.y, GameConst::HEADING_N);
  e.set_int("heading", heading);

  return true;
}

//===================================================================

void Game::move_objects()
{
  PROFILE("Game::move_objects");
  static Array<GameObj*> all_objs;

  {
    PROFILE("zcat check");
    
    all_objs.set_size(0);
    
    // object zcat check and partition
    
    sint4 n = 0;
    FORS (p, player_n+1) {
      ObjCont &obj = objs[p];
      FORALL (obj, i) {
        GameObj *gb = (*i)->get_GameObj(); // check if the ScriptObj is a GameObj
        if (gb == 0) continue;
        ScriptObj::check_ptr(gb);
        GameObj &gob = *gb;
        ++n;
        
        if ((*gob.sod.zcat != Object::ZCAT_UNDEF) &&
            (*gob.sod.x >= 0))
          all_objs.push_back(&gob);
      }
    }
  }
  {
    PROFILE("move objects");

    // move all objects
    // NOTE: This clobbers the contents of motion_secs.
    // The sector will only be valid for the objects last moved in it.
    
    move_objects(all_objs, motion_secs, true);
  }

  // fixme: This might be better done within each move_objects call.
  //        The objects don't necessarily have to move or even have these attributes.
  //        I'm side-stepping this by not enforcing the assignment.

  // (code removed when sod became pointers)
}

//===================================================================

/** Resets object movement and updates motion sectors. */

void Game::update_pos(GameObj &a)
{
  ScriptObj::check_ptr(&a);

  *a.sod.speed = 0;
  *a.sod.is_moving = 0;
  a.smd.fine_heading.x = 0;
  a.smd.fine_heading.y = 0;

  if (*a.sod.x < 0) return; // assumed to be offscreen
  a.update_points();

  Vector<sint4> f;
  compute_ms_footprint(a, f);

  FORALL (f, i) {
    motion_secs.insert(&a, *i);
  }
}

//===================================================================

void Game::inc_magic()
{
  ++magic;
  if (magic == 0) {

    // wrap around: clear all neighbor lists and reset visit value

    FORALL (boundaries, i) {
      GameObj *gob = (*i)->get_GameObj();
      if (gob == 0) continue;
      gob->smd.visit_magic = magic;
      gob->smd.neighbor_magic = magic;
      gob->smd.unique_magic = magic;
    }

    FORS (i, player_n) {
      FORALL (objs[i], it) {
        GameObj *gob = (*it)->get_GameObj();
        if (gob == 0) continue;
        gob->smd.visit_magic = magic;
        gob->smd.neighbor_magic = magic;
        gob->smd.unique_magic = magic;
      }
    }
    ++magic;  // magic != 0
  }
}

//===================================================================

void Game::compute_ms_footprint(const GameObj &a, Vector<sint4> &motion_sectors) const
{
  motion_sectors.clear();

  sint4 s1 = *a.sod.shape;

  sint4 x0,x1,y0,y1;

  if (s1 == Object::CIRCLE) {
    sint4 x = fine_grid(*a.sod.x);
    sint4 y = fine_grid(*a.sod.y);
    sint4 r = fine_grid(*a.sod.radius);
    x0 = x - r;
    x1 = x + r;
    y0 = y - r;
    y1 = y + r;
  } else if (s1 == Object::RECTANGLE || s1 == Object::LINE) {
    x0 = fine_grid(*a.sod.x1);
    y0 = fine_grid(*a.sod.y1);
    x1 = fine_grid(*a.sod.x2);
    y1 = fine_grid(*a.sod.y2);
  } else {
    cout << "type = " << s1 << " : not computing footprint" << endl;
    return;
  }

  x0 /= sec_wh;
  x1 /= sec_wh;
  y0 /= sec_wh;
  y1 /= sec_wh;

  if (x0 < 0) x0 = 0; else if (x1 >= sec_nx) x1 = sec_nx-1;
  if (y0 < 0) y0 = 0; else if (y1 >= sec_ny) y1 = sec_ny-1;

  for(sint4 y = y0; y <= y1; ++y) {
    sint4 s = y*sec_nx + x0;
    for(sint4 x = x0; x <= x1; ++x) {
      motion_sectors.push_back(s);
      ++s;
    }
  }
}

//===================================================================

void Game::compute_ms_footprint(sint4 x0, sint4 y0, sint4 x1, sint4 y1,
                                Vector<sint4> &motion_sectors) const
{
  motion_sectors.clear();

    x0 = fine_grid(x0);
    y0 = fine_grid(y0);
    x1 = fine_grid(x1);
    y1 = fine_grid(y1);

  x0 /= sec_wh;
  x1 /= sec_wh;
  y0 /= sec_wh;
  y1 /= sec_wh;

  if (x0 < 0) 
      x0 = 0; 
  else if (x1 >= sec_nx) 
      x1 = sec_nx-1;
  if (y0 < 0) 
      y0 = 0; 
  else if (y1 >= sec_ny) 
      y1 = sec_ny-1;

  for(sint4 y = y0; y <= y1; ++y) {
    sint4 s = y*sec_nx + x0;
    for(sint4 x = x0; x <= x1; ++x) {
      motion_sectors.push_back(s);
      ++s;
    }
  }
}

//===================================================================

/** Move objects one tick respecting boundaries.
    Important: sod set for each object.

    NOTE: This clobbers the contents of the sector being used.
    The sector will only be valid for the objects last moved in it.

    @bug  Objects will keep "moving" for one more cycle if they reach their
          target location after rounding.

    @param mobjs     Mobile objects (possibly still).
    @param ms        The layer to move the objects on.  Contains static objects (boundaries)
                     and the objects last moved on the layer.
    @param obj_coll  Objects can collide with each other.
*/

void Game::move_objects(Array<GameObj*> &mobjs, MotionSectors &ms, bool obj_coll)
{
  uint4 n = mobjs.size();
  if (n <= 0) return;

  inc_magic();


  // compute motion bounding spheres
  {
    PROFILE ("bounding spheres");

    FORU (i, n) {
      
      GameObj &e = *mobjs[i];
      ScriptObj::check_ptr(&e);
      if (e.is_dead()) continue;
      assert(!e.sod.is_boundary); // boundaries shouldn't be part of the moving objects
      
      if (*e.sod.is_moving) {
        
        // last tick => adjust heading such that target is reached with last step
        if (e.smd.ticks_so_far + 1 >= e.smd.ticks_total) {
          ScalarPoint delta = e.smd.target_f_pos;
          delta.sub(e.smd.fine_pos);
          e.smd.fine_heading = delta;
        }
        
        // moving object: motion sphere
        // (-x-)-->--M----(-x-)
        
        ScalarPoint h = e.smd.fine_heading;
        h.scale(0.5);
        
        e.smd.bc_fine_center = e.smd.fine_pos;
        e.smd.bc_fine_center.add(h);
        
        e.smd.bc_fine_r = FixedPoint::add(h.l2_norm(), fine_grid(*e.sod.radius) + 2);
        // + 2: conservative error margin for rounding
        
      } else {
        
        *e.sod.speed = 0;
        
        // still object: bounding sphere
        
        if (*e.sod.shape == Object::RECTANGLE ||
            *e.sod.shape == Object::LINE) {
          
          e.update_points();
          
          e.smd.bc_fine_center.x = fine_grid(*e.sod.x);
          e.smd.bc_fine_center.y = fine_grid(*e.sod.y);

	  Scalar diag =
	    (Scalar)(sqrt((real8)(square((sint4)(*e.sod.x1 - *e.sod.x2))+
			  square((sint4)(*e.sod.y1 - *e.sod.y2)))
			  ));
	  
	  e.smd.bc_fine_r = fine_grid(diag/2+1);
          
        } else { // circle
          
          e.smd.bc_fine_center.x = fine_grid(*e.sod.x);
          e.smd.bc_fine_center.y = fine_grid(*e.sod.y);
          e.smd.bc_fine_r        = fine_grid(*e.sod.radius);
        }
      }
    }
  }

  uint1* &active = ms.active;
  Array<sint4, 600> &active_sectors = ms.active_sectors;

  {
    PROFILE ("fill motion sectors");

    // determine sector membership
    
    FORS (it, active_sectors.size()) { // all active sectors ...
      sint4 as = active_sectors[it];
      MotionSectorInfo &si = ms.secs[as];
      si.set_size(ms.sizes[as]); // reset the sizes (remove objects from previous cycle)
      // they were kept around to speed up collision testing for the actions
    }
    active_sectors.set_size(0);
    memset(active, 0, sec_num*sizeof(active[0]));
    
    // compute which motion sectors the boundary circles (computed above) intersect
    
    FORU (i, n) {
      
      GameObj &e = *mobjs[i];
      if (e.is_dead()) continue;
      
      e.smd.neighbors.set_size(0); // boundaries done with ... magic
      
      sint4 x = e.smd.bc_fine_center.x;
      sint4 y = e.smd.bc_fine_center.y;
      sint4 r = e.smd.bc_fine_r;
      
      sint4 x0 = (x - r);
      sint4 x1 = (x + r);
      sint4 y0 = (y - r);
      sint4 y1 = (y + r);
      
      x0 /= sec_wh;
      x1 /= sec_wh;
      y0 /= sec_wh;
      y1 /= sec_wh;
      
      if (x0 < 0) x0 = 0; else if (x1 >= sec_nx) x1 = sec_nx-1;
      if (y0 < 0) y0 = 0; else if (y1 >= sec_ny) y1 = sec_ny-1;
      
      // insert the object into the sectors it (could) intersect
      for (sint4 y = y0; y <= y1; ++y) {
        sint4 s = y*sec_nx + x0;
        for (sint4 x = x0; x <= x1; ++x) {
          ms.insert(&e, s);
          ++s;
        }
      }
    }
  }

  {
    PROFILE("potential pairs");

    // generate all pairs of elements that potentially collide
    
    FORS (it, active_sectors.size()) { // all active sectors ...
      
      sint4 as = active_sectors[it];
      MotionSectorInfo &si = ms.secs[as];
      
      //cout << "sec " << active_sectors[it] << " " << si.size() << endl;
      
      // Boundaries are at the beginning of the array, objects are after.
      // Use this to speed up finding collision pairs by modifying the range of the
      // outside loop based on whether objects are colliding with each other (obj_coll).
      
      sint4 boundary_end = ms.sizes[as];
      sint4 object_end = si.size();
      
      // don't need to loop over objects in the outer loop if ignoring object collisions
      sint4 sec_end;
      if (obj_coll) sec_end = object_end;
      else          sec_end = boundary_end;
      
      
      FORS (i1, sec_end) {
        
        GameObj &e1 = *si[i1];
        
        // don't include boundaries in this loop
        for (sint4 i2 = max(boundary_end, i1+1); i2 < object_end; ++i2) {
          
          GameObj &e2 = *si[i2];
          if (!*e1.sod.is_moving && !*e2.sod.is_moving) continue;
          
          Scalar rsum = e1.smd.bc_fine_r + e2.smd.bc_fine_r;
          
          // do motion spheres intersect?
          if (e1.smd.bc_fine_center.max_dist(e2.smd.bc_fine_center) <= rsum) { // infinity norm
            
            // actual distance
            Scalar dist = e1.smd.bc_fine_center.l2_dist(e2.smd.bc_fine_center);
            
            // circles intersect (potential collision)
            if (dist <= rsum) {
              
#if TEST_OUT
              cout << "EDGE (" << i1 << "," << i2 << ")" << endl;
#endif
              if (e1.smd.neighbor_magic != magic) {
                e1.smd.neighbor_magic = magic; e1.smd.neighbors.set_size(0);
              }
              if (e2.smd.neighbor_magic != magic) {
                e2.smd.neighbor_magic = magic; e2.smd.neighbors.set_size(0);
              }
              e1.smd.neighbors.push_back(&e2);
              e2.smd.neighbors.push_back(&e1);
            }
          }
        }
      }
    }
  }

  const sint4 OBJ_NMAX = boundaries.size()+n; // boundary sectors + moving objs!
  GameObj* *comp = new GameObj*[OBJ_NMAX];

  {
    PROFILE("handle components");

    // handle movement in connected components
    
    visit_magic = magic;
    
    FORU (i, n) {
      
      CycleCounter start_time;
      GameObj **p = comp;

      bool mobile = false;
      {
        PROFILE("visit");

        // determine if a group of objects has a moving member
        visit(*mobjs[i], p); // Construct a network of connected objects (uses magic!)
        for (GameObj **q = comp; q < p; ++q) {
          ScriptObj::check_ptr(*q);
          if (*(*q)->sod.is_moving) {
            mobile = true;
            break;
          }
        }
      }
      
      if (p-comp > OBJ_NMAX) ERR("p-comp > OBJ_NMAX");
      
      
      if (p != comp && mobile) {
        PROFILE("handle_component");
        //      cout << "A= " << p << " " << comp << " " << (p-comp) << " " << n << " " << OBJ_NMAX << endl;
        handle_component(comp, p-comp);
      }
    }
  }
  
  delete [] comp;
}

//===================================================================

/** Reposition objects on the client.  Place objects into the sectors they intersect,
    since no motion functions are being called to do this automatically.
*/

void Game::update_sectors()
{
  static Array<GameObj*> all_objs;
  all_objs.set_size(0);

  // object zcat check and partition

  FORS (p, player_n+1) {
    ObjCont &obj = objs[p];
    FORALL (obj, i) {
      GameObj *gob = (*i)->get_GameObj();
      if (gob == 0) continue;

      if ((*gob->sod.zcat != Object::ZCAT_UNDEF) &&
          (*gob->sod.x >= 0))
        all_objs.push_back(gob);
    }
  }

  sint4 n = all_objs.size();

  if (n <= 0) return;

  Array<GameObj*> &mobjs = all_objs;
  MotionSectors &ms = motion_secs;

  uint1* &active = ms.active;
  Array<sint4, 600> &active_sectors = ms.active_sectors;

  memset(active, 0, sec_num*sizeof(active[0]));

  FORS (it, active_sectors.size()) {
    sint4 as = active_sectors[it];
    MotionSectorInfo &si = ms.secs[as];
    si.set_size(ms.sizes[as]); // reset the sizes (remove objects from previous cycle)
    // they were kept around to speed up collision testing for the actions
  }
  active_sectors.set_size(0);

  FORS (i, n) {

    GameObj &e = *mobjs[i];
    if (e.is_dead()) continue;
    //    e.update_points();

    Vector<sint4> ms_footprint;
    compute_ms_footprint(e, ms_footprint);

    FORALL (ms_footprint, i) {
      sint4 s = *i;
      assert(s >= 0 && s < sec_num);
      if (!active[s]) {
        active[s] = true;
        active_sectors.push_back(s);
      }
      ms.secs[s].push_back(&e);
    }
  }
}

//===================================================================

// faster: update collision times in vicinity of colliding elements only

struct Edge {

public:

  GameObj *o1, *o2;    // o1 < o2
  sint4 coll_type; // collision type returned by the collision test,
                   // used to resolve stopping objects later

  // used to assign 1 to o2 per default 
  Edge(GameObj *o1_=0, GameObj *o2_=0, sint4 coll_type_=0) :
    o1(o1_), o2(o2_), coll_type(coll_type_)
  {
    //    if (o1 == o2) ERR("LOOP");
    if (o1 > o2) { GameObj *t = o1; o1 = o2; o2 = t; }
  }

  inline bool operator < (const Edge &e) const {
    if (o1 < e.o1) return true;
    if (o1 > e.o1) return false;
    return o2 < e.o2;
  }

  inline bool operator == (const Edge &e) const {
    if (o1 != e.o1) return false;
    if (o2 != e.o2) return false;
    return true;
  }

  inline bool operator != (const Edge &e) const {
    if (o1 != e.o1) return true;
    if (o2 != e.o2) return true;
    return false;
  }
};


#if 1

// sint4 seems to be sufficient

class CollTime {

public:

  sint4 v;

#define CT_FAC 100000000L  // 10^8
#define CT_FAC_INV (1.0L/CT_FAC)

  CollTime() { v = 0; }
  void from_real8(real8 f) { v = sint4(f * CT_FAC); }
  real8 to_real8() const { return real8(v) * CT_FAC_INV; }

  bool operator == (const CollTime &c) const { return v == c.v; }
  bool operator <  (const CollTime &c) const { return v < c.v; }
  bool operator >  (const CollTime &c) const { return v > c.v; }

};

#else

class CollTime {

public:

  sint8 v;

#define CT_FAC 1000000000000LL  // 10^12

  CollTime() { v = 0LL; }
  void from_real8(real8 f) { v = sint8(f * CT_FAC); }
  real8 to_real8() const { return real8(v)/CT_FAC; }

  bool operator == (const CollTime &c) const { return v == c.v; }
  bool operator <  (const CollTime &c) const { return v < c.v; }

};

#endif

struct Edge2 {

public:

  GameObj *o1, *o2;    // o1 < o2
  CollTime ct;
  sint4 coll_type; // collision type returned by the collision test,
                   // used to resolve stopping objects later

  // used to assign 1 to o2 per default   
  Edge2(CollTime ct_, GameObj* o1_=0, GameObj* o2_=0, sint4 coll_type_=0) {
    coll_type = coll_type_;
    o1 = o1_; o2 = o2_; ct = ct_;
    //    if (o1 == o2) ERR("LOOP");
    if (o1 > o2) { GameObj* t = o1; o1 = o2; o2 = t; }
  }

  inline bool operator < (const Edge2 &e) const {
    if (ct < e.ct) return true;
    if (ct > e.ct) return false;
    if (o1 < e.o1) return true;
    if (o1 > e.o1) return false;
    return o2 < e.o2;
  }
};

//===================================================================

const bool FULL_STOP = 0; // stop everyone involved in a collision even if not blocked

void Game::handle_component(GameObj** comp, sint4 n)
{
  if (n <= 0) return;

#if TEST_OUT
  cout << "------------ COMPONENT ------------" << endl;

  { FORS (i, n) { cout << comp[i] << " "; } }
  cout << endl;
#endif

  set<Edge2> eh2;
  PairHeap<Edge, CollTime> eh;

  GameObj* *stopped = new GameObj*[n+1];
  GameObj* *new_stopped = new GameObj*[n+1];

  GameObj** p_stopped;
  GameObj** p_new_stopped;
  sint4 stopped_n, new_stopped_n;

  // compute all collision times

  p_stopped = stopped;

  // Run through all the objects in the component and compute collision times for
  // each pair.  Resolve the effects of local collisions in the next block.

  FORS (i, n) {

    inc_magic();
    GameObj &e1 = *comp[i];
    const Neighbors &ninf = e1.smd.neighbors;

    // fixme: Check if this block is needed to correct for reaching the target in the middle of a tick.
#if 0
    if (*e1.sod.is_moving && (e1.smd.ticks_so_far + 1 >= e1.smd.ticks_total)) {
      CollTime ct;
      real8 t = 1.0 / e1.smd.inv_ttt - e1.smd.ticks_so_far;
      ct.from_real8(t);
      Edge e(&e1, &e1, 0);
      
      eh.ins(e, ct);
    }
#endif

    FORS (ni, ninf.size()) {

      GameObj &e2 = *ninf[ni];

      if (&e1 >= &e2) continue;                   // wrong order
      if (e2.smd.unique_magic == magic) continue; // already seen
      e2.smd.unique_magic = magic;

      if (*e1.sod.is_moving || *e2.sod.is_moving) {

        sint4 coll_type;
        real8 ctf = collision_time(e1, e2, 0.0, coll_type);

#if TEST_OUT
        cout << "CHECK " << &e1 << " " << &e2 << " CT " << ctf << endl;
#endif
        if (ctf < 0.0) TOO_CLOSE("A", &e1, &e2);

        if (ctf == 0.0) {

          //          cout << "immediate stop" << endl;

          // immediate stop
          // last parameter is unimportant for circles
          sint4 objs_stopped = resolve_collision(e1, e2, ctf, coll_type);

          if (*e1.sod.is_moving && (objs_stopped & 1 || FULL_STOP)) {
            obj_stop(e1);
            *p_stopped++ = &e1;
          }
          if (*e2.sod.is_moving && (objs_stopped & 2 || FULL_STOP)) {
            obj_stop(e2);
            *p_stopped++ = &e2;
          }

#if TEST_OUT
          cout << "STOPPED(0-1) " << &e1 << " " << &e2 << endl;
#endif

        } else if (ctf <= 1.0) {

          CollTime ct;
          ct.from_real8(ctf);
          Edge e(&e1, &e2, coll_type);
          
          eh.ins(e, ct);

        }
      }
    }
  }

  CollTime curr_time;

  stopped_n = p_stopped - stopped;
  p_stopped = stopped;
  p_new_stopped = new_stopped;

  // Keep applying the earliest collision and recalculating collision times until no more occur.

  FOREVER {

    while (stopped_n > 0) { // while objects are still colliding

      // update all collision times in the direct vicinity of stopped elements

      new_stopped_n = 0;

      FORS (i, stopped_n) {

        GameObj &e1 = *p_stopped[i];
        Neighbors &ninfo = e1.smd.neighbors;

        FORS (ni, ninfo.size()) {

          GameObj &e2 = *ninfo[ni];
          ScriptObj::check_ptr(&e2);

          if (*e2.sod.is_moving) {

            real8 ctf = curr_time.to_real8();

            // check this:           // any collision would have to stop e2
            sint4 coll_type;
            real8 ct = collision_time(e1, e2, ctf, coll_type);
            real8 diff = ct - ctf;
            Edge e(&e1, &e2, coll_type);

            //            if (diff < 0.0) TOO_CLOSE("A", &e1, &e2);

            if (diff <= 0.0) {

              //              cout << "immediate stop 2" << endl;

              sint4 objs_stopped = resolve_collision(e1, e2, ct, coll_type);

              if (objs_stopped & 2 || FULL_STOP) {
                obj_move_and_stop(e2, ctf);
                p_new_stopped[new_stopped_n++] = &e2;
              }

#if TEST_OUT
              cout << "STOPPED(0-2) " << p_stopped[i]
                   << " " << &e2 << " " << ctf << endl;
#endif

              // remove dead edge
              eh.del(e);

            } else if (ct >= 0.0 && ct <= 1.0L) {

              // update collision time
              CollTime nct;
              nct.from_real8(ct);
              eh.ins(e, nct);

#if TEST_OUT
              cout << ">>>>>>>>>>>>>> UPDATE " << e.o1 << " " << e.o2 << " " << ctf << " " << diff << endl;
#endif
            } else {

              // no collision anymore
              CollTime nct;
              nct.from_real8(2.0);
              //              eh.ins(e, ct);

            }

          } else {

            // remove dead edge if it still exists
            Edge e(&e1, &e2, 0);
            eh.del(e);
          }
        }
      }

      GameObj** t = p_stopped;
      p_stopped = p_new_stopped;
      p_new_stopped = t;


      stopped_n = new_stopped_n;
    }


    bool finished = false;

    FOREVER {
      
      Edge e;
      
      FOREVER {
        if (!eh.min(e, curr_time) || curr_time.to_real8() > 1.0) {
          finished = true; break;
        }
        // This check is now needed because of o1 == o2 "end of step" pseudo-collisions.
        // Otherwise stationary pairs should have been removed already.
        if (*e.o1->sod.is_moving || *e.o2->sod.is_moving) break;
        eh.del(e);
      }
      
      if (finished) break;
      
      real8 ctf = curr_time.to_real8();
      stopped_n = 0;
      
      sint4 objs_stopped = resolve_collision(*e.o1, *e.o2, ctf, e.coll_type);
      
      if (*e.o1->sod.is_moving && (objs_stopped & 1 || FULL_STOP)) {
        obj_move_and_stop(*e.o1, ctf);
        stopped[stopped_n++] = e.o1;
      }
      if (*e.o2->sod.is_moving && (objs_stopped & 2 || FULL_STOP)) {
        obj_move_and_stop(*e.o2, ctf);
        stopped[stopped_n++] = e.o2;
      }


      if (stopped_n == 0) {
        eh.del(e);
        continue;
      }
      break;
    }

    if (finished) break;


#if TEST_OUT
    cout << "STOPPED XXX " << e.o1 << " " << e.o2 << " " << ctf << endl;
#endif

    p_stopped = stopped;
    p_new_stopped = new_stopped;
  }

  delete [] stopped;
  delete [] new_stopped;


  // remaining moves; no collisions left to check, free movement until end of cycle

  FORS (i, n) {

    real8 time_step = 1.0;
    GameObj &e = *comp[i];

    if (*e.sod.is_moving) {

#ifndef NDEBUG
      if (e.sod.is_boundary) ERR("WHOOPS");
#endif

      obj_move(e, time_step);
      obj_move_to_grid(e);

#if TEST_OUT
      cout << "NON_COLL MOVED " << comp[i];
      cout << " P="; e.smd.fine_pos.write();
      cout << endl;
#endif

      e.smd.ticks_so_far++;
      if (e.smd.ticks_so_far >= e.smd.ticks_total) {
        //        *e.sod.speed = 0;
        *e.sod.is_moving = 0;
      }
    }
  }

  //  t_heap += eh.ticks; // timing data
}

//===================================================================
