#ifndef Net_Protocol_H
#define Net_Protocol_H

// $Id: Net_Protocol.H,v 1.6 2006/06/29 01:04:59 orts_mburo Exp $

#include "Net_Message.H"
#include <sstream>


// message type strings
static const std::string  INVITE       = "invite";
static const std::string  INVIT_DISC   = "invitdisc";    

static const std::string  ABORT_       = "abort";        
static const std::string  ABORTED      = "aborted";      
static const std::string  KICK         = "kick";         
static const std::string  KICKED       = "kicked";        
static const std::string  JOIN         = "join";          
static const std::string  JOINED       = "joined";        
static const std::string  LEAVE        = "leave";         
static const std::string  LEFT         = "left";          
static const std::string  READY        = "ready";         
static const std::string  READIED      = "readied";       
static const std::string  UNREADY      = "unready";       
static const std::string  UNREADIED    = "unreadied";     
static const std::string  GAME_START   = "gstart";       
static const std::string  GAME_INFO    = "ginfo";
static const std::string  WORKER_START = "wstart";       
static const std::string  GAME_END     = "game_end";      
static const std::string  WORKER_READY = "worker_ready";  
static const std::string  ERROR        = "error";        
static const std::string  GGS_ERROR        = "ERR";        

static const std::string  INVITATIONS  = "invitations";   
static const std::string  GAMES        = "games";         
static const std::string  WORKERS      = "workers";       
static const std::string  BUDDIES      = "buddies";       

static const std::string  TELL         = "tell";          

static const std::string  BUDDY        = "buddy";         
static const std::string  LOGIN_NOTIFY = "login_n";       
static const std::string  ADDED_BUDDY  = "added_buddy";   
static const std::string  REMED_BUDDY  = "remed_buddy";   

// list names used to request a list from the bot
static const std::string  LIST         = "list";         
  static const std::string  BUDDY_LIST   = "buddy_list";
  static const std::string  WHO_LIST     = "who_list";
  static const std::string  GAME_LIST    = "game_list";
  static const std::string  INVITE_LIST  = "invite_list";
  static const std::string  PLAYER_LIST  = "players_list";

static const std::string  NOTIFY       = "notify";        
static const std::string  REGISTER     = "register";     
static const std::string  LOGIN        = "login";        
static const std::string  DOWN         = "down";         
static const std::string  FINGER       = "finger";

static const std::string  WORKER_AVAILABLE = "worker_available";

/** the levels one can log on the the bot with */
typedef enum Level_{ ROOT = 0, SUPER = 1, GENERAL = 2, WORKER = 3 }Level;

namespace Protocol{
  class  Message_Base;
  class  Invite;
  class  Invit_Disc;
  class  Abort;
  class  Aborted;
  class  Kick;
  class  Kicked;
  class  Join;
  class  Joined;
  class  Leave;
  class  Left;
  class  Ready;
  class  Readied;
  class  Unready;
  class  Unreadied;
  class  Game_Start;
  class  Game_Info;
  class  Worker_Start;
  class  Worker_Ready;
  class  Game_End;
  class  Invitations;
  class  Games;
  class  Workers;
  class  Tell;
  class  Buddy;
  class  Login_Notify;
  class  Added_Buddy;
  class  Remed_Buddy;
  class  Buddies;
  class  Error;
  class  GGS_Error;
  class  Player_List;
  class  Who_List;
  
  class  GGS_Notify;
  class  Register;
  class  Login;
  class  Down;
  class  Finger;
  class  List;
  
  class  Worker_Available;
};

using namespace Protocol;


/***************************************************************************************************
 * @class Message_Base
 * contains info and functions common to all messages
 ***************************************************************************************************/
class Protocol::Message_Base{
public:
  virtual bool parse(const Net_Message &m) = 0;
  virtual ~Message_Base(){}
  std::string get_sender() const { return sender; }
  std::string get_message() const { return message;}/* get the message type */

protected:

  /**
    + = true
    - = false
    otherwise return false
  */
  bool set_bool(const std::string& message, bool& toSet);

  /**
    if message == "NULL" return false
    otherwise set the std::string
  */
  bool set_string(const std::string& message, std::string& toSet);

  /**
    if b is true return "+" otherwise "-"
  */
  static std::string bool_to_string(bool b);
  static std::string to_string(int a);

  static bool set_int(const std::string &astr, int& anInt);

  /**
    replace all spaces with '_' so that this std::string is not broken up be parser
  */
  static void solidify(std::string& message);

  /**
    replace all '_' with space
  */
  static void melt(std::string& message);

  std::string message;
  std::string sender;
};

/***************************************************************************************************
 * @class Invite
 * sent to the bot to create an invitation
 * <game_name> <maxPlayers> <invisible(+|-)> <anonymous (+|-)> <anonymousid> <pass> <options>
 ***************************************************************************************************/
class Protocol::Invite:public Protocol::Message_Base{
public:

  virtual ~Invite(){}
  bool parse(const Net_Message &m);

  std::string get_game_name()   const { return name; }
  int get_nplayers()       const { return nplayers;}
  bool get_invisible()     const { return invisible; }
  bool get_anony_ids()     const { return anony_ids; }
  bool get_passprotected() const { return passprotected; }
  std::string get_alias()       const { return alias; }
  std::string get_password()    const { return password; }
  std::string get_options()     const { return options;   }

  static std::string compose(const std::string &name,
                        const int numPlayers,
                        const bool invisible,
                        const bool anony,
                        const bool passprotected,
                        const std::string& anonyname,
                        const std::string& pass,
                        const std::string& options);

protected:
  std::string name;
  int nplayers;
  bool invisible;
  bool anony_ids;
  bool passprotected;
  std::string alias;
  std::string password;
  std::string options;
};




/************************************************************************************************
 * an invitations discriptor class used to send out and parse currently running game invitations
 * <iid> <nplayers> <owner> <game_name> <maxPlayers> <invisible(+|-)> <anonymous (+|-)> <pass> <options>
 *************************************************************************************************/
class Protocol:: Invit_Disc:public Protocol::Message_Base{
public:

  virtual ~Invit_Disc(){}
  bool parse(const Net_Message &m);

  int     get_iid()         const { return iid;        }
  int     get_num_players() const { return nPlayers;   }
  std::string  get_owner()       const { return owner;      }
  std::string  get_game_name()   const { return gameName;   }
  int     get_max_players() const { return maxPlayers; }
  bool    get_invisible()   const { return invisible;  }
  bool    get_anony_ids()   const { return anony_ids;  }
  bool    get_passprotected() const { return passprotected; }
  std::string  get_options()     const { return options;    }

  static std::string compose(const int iid,
                        const int nPlayers,
                        const std::string& owner,
                        const std::string& gameName,
                        const int maxPlayers,
                        const bool invisible,
                        const bool anony,
                        const bool passprotected,
                        const std::string& options);

protected:

  int iid;
  int nPlayers;
  std::string owner;
  std::string gameName;
  int maxPlayers;
  bool invisible;
  bool anony_ids;
  bool passprotected;
  std::string options;

};




/***************************************************************************************************
 * @class Abort
 * sent to the bot to abort an invitation
 * <iid>
 ***************************************************************************************************/
class Protocol:: Abort:public Protocol::Message_Base{
public:
  virtual ~Abort(){}

  bool parse(const Net_Message &m);
  static std::string compose(int id);

  //accessors
  int get_iid() const { return iid; }
  
protected:
  int iid;
};




/***************************************************************************************************
 * @class Aborted
 * sent out by the bot to all players to notify them that a game has been aborted
 * <iid>
 ***************************************************************************************************/
class Protocol:: Aborted:public Protocol::Message_Base{
public:
  virtual ~Aborted(){}

  bool parse(const Net_Message &m);
  static std::string compose(int id);

  //accessors
  int get_iid() const { return iid; }
  
protected:
  int iid;
};




/***************************************************************************************************
 * @class Kick
 * sent by the owner of and invitation to the bot to kick a player out of an invitation
 * <iid> <pid>
 ***************************************************************************************************/
class Protocol:: Kick:public Protocol::Message_Base{

public:
  virtual ~Kick(){}

  bool parse(const Net_Message &m);
  static std::string compose(int iid, int pid);

  // accessors
  int get_pid() const { return pid; }
  int get_iid() const { return iid; }
  
protected:
  int pid;
  int iid;
};




/***************************************************************************************************
 * @class Kicked
 * when a player is kick out from an invitation this is sent out by the bot to all the other players
 * in that game
 * <iid> <pid>
 ***************************************************************************************************/
class Protocol:: Kicked:public Protocol::Message_Base{
public:
  virtual ~Kicked(){}
  bool parse(const Net_Message &m);
  static std::string compose(int iid, int pid);

  // acessors
  int get_pid() const { return pid; }
  int get_iid(){ return iid; }
  
protected:
  int pid;
  int iid;
};




/***************************************************************************************************
 * @class Join
 * sent by a player to the bot to join a specific game
 * <iid> [pass]
 ***************************************************************************************************/
class Protocol:: Join:public Protocol::Message_Base{
public:
  virtual ~Join(){}
  bool parse(const Net_Message &m);
  static std::string compose(int id,
                        bool anony_ids,
                        const std::string &alias,
                        bool passprotected,
                        const std::string& pass = "NULL");

  // accessors
  int get_iid() const { return iid; }
  bool get_anony_ids() const { return anony_ids; }
  bool get_passprotected() const { return passprotected; }
  std::string get_alias() const { return alias; }
  std::string get_pass() const{ return pass; }
  
protected:
  int iid;
  bool anony_ids;
  bool passprotected;
  std::string alias;
  std::string pass;
};




/***************************************************************************************************
 * @class Joined
 * set out by the bot to all players in an invitation when a new player joins that invitation
 * <iid> <pid>
 ***************************************************************************************************/
class Protocol:: Joined:public Protocol::Message_Base{
public:
  virtual ~Joined(){}

  bool parse(const Net_Message &m);
  static std::string compose(int gid, int pid, bool anony, const std::string& alias);

  //accessors
  int  get_iid() const { return iid; }
  int  get_pid() const { return pid; }
  bool get_anony() const { return pid; }
  const std::string&  get_player_alias() const { return player_alias;}
  
protected:
  int iid;
  int pid;
  bool anony;
  std::string player_alias;
  
};




/***************************************************************************************************
 * @class Leave
 * sent by a player to the bot to leave a particular invitation
 * <iid>
 ***************************************************************************************************/
class Protocol:: Leave:public Protocol::Message_Base{
public:
  virtual ~Leave(){}
  bool parse(const Net_Message &m);
  static std::string compose(int id);

  // accessors
  int get_iid() const { return iid; }

protected:
  int iid;
};




/***************************************************************************************************
 * @class Left
 * when a player leaves this is sent out to all other players in that invitation
 * <iid> <pid>
 ***************************************************************************************************/
class Protocol:: Left:public Protocol::Message_Base{
public:
  virtual ~Left(){}
  bool parse(const Net_Message &m);
  static std::string compose(int id, int pa);

  //accessors
  int get_iid() const { return iid; }
  int get_pid() const { return pid;}

protected:
  int pid;
  int iid;
};



/***************************************************************************************************
 * @class Ready
 * sent to the bot by a player to set their ready status to true
 * <iid>
 ***************************************************************************************************/
class Protocol:: Ready:public Protocol::Message_Base{
public:
  virtual ~Ready(){}
  int get_iid() const { return iid; }
  static std::string compose(int id);

  //accessors
  bool parse(const Net_Message &m);
  
protected:
  int iid;
};





/***************************************************************************************************
 * @class Readied
 * sent out by the bot to all players in an invitation when one of the players sets their status to
 * ready
 * <iid> <pid>
 ***************************************************************************************************/
class Protocol:: Readied:public Protocol::Message_Base{
public:
  virtual ~Readied(){}
  bool parse(const Net_Message &m);
  static std::string compose(int iid, int pid);

  // accessors
  int get_iid() const { return iid;}
  int get_pid() const { return pid;}
  
protected:
  int pid;
  int iid;
};




/***************************************************************************************************
 * @class Unready
 * sent to the bot by a player to set their ready status to true
 * <iid>
 ***************************************************************************************************/
class Protocol:: Unready:public Protocol::Message_Base{
public:
  virtual ~Unready(){}
  bool parse(const Net_Message &m);
  static std::string compose(int id);

  //accessors
  int get_iid() const { return iid; }
  
protected:
  int iid;
};



/***************************************************************************************************
 * @class Unreadied
 * sent out by the bot to all players in an invitation when one of the players sets their status to
 * unready
 * <iid> <pid>
 ***************************************************************************************************/
class Protocol:: Unreadied:public Protocol::Message_Base{
public:
  virtual ~Unreadied(){}
  bool parse(const Net_Message &m);
  int get_iid() const { return iid; }
  int get_pid() const { return pid;}
  static std::string compose(int iid, int pid);
protected:
  int pid;
  int iid;
};




/***************************************************************************************************
 * @class Game_Start
 * set out by the bot to tell all players in an invitation to connect to the worker
 * no params
 ***************************************************************************************************/
class Protocol:: Game_Start:public Protocol::Message_Base{

public:

  virtual ~Game_Start(){}
  bool parse(const Net_Message& m);
  static std::string compose();

  protected:
};



/***************************************************************************************************
 * @class Game_Info
 * sent out by the bot. contains information about the worker
 * creates an options std::string
 * ie -shost foo -netpass bar
 * <shost> <netpass>
 ***************************************************************************************************/
class Protocol:: Game_Info:public Protocol::Message_Base{

public:

  virtual ~Game_Info(){}
  bool parse(const Net_Message& m);
  static std::string compose(const std::string& shost, const std::string& netpass);
  std::string get_info () const { return info; }

protected:
  std::string info;
};



/***************************************************************************************************
 * @class Worker_Start
 * sent by the bot to the worker to tell it to start the game server and start listening for
 * connections
 * <numPlayers> <options> (<player1IP> <player1pass>)*...
 ***************************************************************************************************/
class Protocol:: Worker_Start:public Protocol::Message_Base{
public:
  virtual ~Worker_Start(){}
  bool parse(const Net_Message &m);

  std::string get_options(){return options;}
  Vector<std::string> get_ipv(){return ipv;}
  Vector<std::string> get_passv(){return passv;}
  int get_num_players(){return numPlayers;}

  static std::string compose(std::string options, int numPlayers, Vector<std::string> ipv, Vector<std::string> passv);
protected:
  int numPlayers;        /** the number of players */
  std::string options;        /** commad line options   */
  Vector<std::string> ipv;    /** Vector if ips that will connect */
  Vector<std::string> passv;  /** the password corresponding to each IP */
};



/***************************************************************************************************
 * @class Worker_Ready
 * sent by the worker to the bot when the server is ready and is listening for connections
 * no params
 ***************************************************************************************************/
class Protocol:: Worker_Ready:public Protocol::Message_Base{
public:
  virtual ~Worker_Ready(){}
  bool parse(const Net_Message &m);
  static std::string compose();

};



/***************************************************************************************************
 * @class Game_End
 * not used yet
 ***************************************************************************************************/
class Protocol:: Game_End:public Protocol::Message_Base{
public:
  virtual ~Game_End(){}
  bool parse(const Net_Message &m);
  static std::string compose(int g_id);

  int get_iid(){return iid;}
  
protected:
  int iid;
};



/***************************************************************************************************
 * @class Invitations
 * a list containing the current invitations. Sent out by the bot upon request
 * <Inv_Discritpor>
 * @see Inv_Discriptor
 ***************************************************************************************************/
class Protocol:: Invitations:public Protocol::Message_Base{/** still not complete */

public:

  virtual ~Invitations(){}

  bool parse(const Net_Message &m);
  
  static std::string compose(const std::string& list);

  //uint4 size() const {  return list.size(); }
  const Invit_Disc& get() const {  return list; }
  
protected:
  Invit_Disc list;
};




/***************************************************************************************************
 * @class Games
 * Used to send out a lset of games
 * not used yet
 ***************************************************************************************************/
class Protocol:: Games:public Protocol::Message_Base{/** still not complete */
public:
  virtual ~Games(){}
  bool parse(const Net_Message &m);
  static std::string compose(const std::string &m);
protected:

};



/***************************************************************************************************
 * @class Workers
 * used to send out a list of workers
 * not used yet
 ***************************************************************************************************/
class Protocol:: Workers:public Protocol::Message_Base{/** still not complete */
public:
  virtual ~Workers(){}
  bool parse(const Net_Message &m);
  static std::string compose(const std::string &m);
protected:

};



/***************************************************************************************************
 * @class Tell
 * sent by a player in an invitation to the bot for telling something to all players in that
 * invitation
 * <message>
 ***************************************************************************************************/
class Protocol:: Tell:public Protocol::Message_Base{
public:
  virtual ~Tell(){}
  bool parse(const Net_Message &m);
  static std::string compose(int iid, int pid, const std::string &outm);

  //accessors
  int get_iid() const { return iid; }
  int get_pid() const { return pid; }
  std::string get_out_msg() const { return out_msg; }

protected:
  std::string out_msg;
  int iid;
  int pid;
};



/***************************************************************************************************
 * @class Error
 * sent by the bot to any client when an error occures
 * <message>
 ***************************************************************************************************/
class Protocol:: Error:public Protocol::Message_Base{
public:
  virtual ~Error(){}
  bool parse(const Net_Message &m);
  static std::string compose(const std::string &msg);

  //accessors
  std::string get_msg() const { return msg; }

protected:
  std::string msg;
};

/***************************************************************************************************
 * @class GGS_Error
 * sent by GGS when an error occures.
 * <message>
 ***************************************************************************************************/
class Protocol:: GGS_Error:public Protocol::Message_Base{
public:
  virtual ~GGS_Error(){}
  bool parse(const Net_Message &m);

  //accessors
  std::string get_msg() const { return msg; }

protected:
  std::string msg;
};



/***************************************************************************************************
 * @class Buddy
 * sent by a player to the bot to add/remove people from their buddy list
 * <+|-> <buddy_name>
 ***************************************************************************************************/
class Protocol:: Buddy:public Protocol::Message_Base{
public:
  virtual ~Buddy(){}
  bool parse(const Net_Message &m);

  /**
    @return true if the player is to be added false otherwise
  */
  bool get_add(){return add;}
  std::string get_buddy_name(){ return buddy_name; }

  static std::string compose(bool add, const std::string &buddyName);

protected:
  std::string buddy_name;
  bool   add;
};




/***************************************************************************************************
 * @class Login_Notify
 * sent out by the bot to all logged on players one someone comes online or goes offline
 * <+|-> <playerName>
 ***************************************************************************************************/
class Protocol:: Login_Notify:public Protocol::Message_Base{
public:
  virtual ~Login_Notify(){}
  bool parse(const Net_Message &m);

  /**
    @return true if the player is online false otherwise
  */
  bool get_online() const {return online;}
  std::string get_player_name() const { return player_name; }

  static std::string compose(bool online, const std::string &playerName);

protected:
  std::string player_name;
  bool online;
};




/***************************************************************************************************
 * @class Added_Buddy
 * sent by the bot to a player to confirm that the buddy as been added and wether he/she is online
 * or offline
 * <player_name> <+|->
 ***************************************************************************************************/
class Protocol:: Added_Buddy:public Protocol::Message_Base{
public:
  virtual ~Added_Buddy(){}
  
  bool parse(const Net_Message &m);
  static std::string compose(const std::string &pn, bool online);

  //accessors
  std::string get_player_name() const { return player_name; }
  bool   get_online() const { return online;}
  
protected:
  std::string player_name;
  bool online;
};



/***************************************************************************************************
 * @class Remmed_Buddy
 * sent by the bot to a player to confirm that the buddy as been removed
 * <player_name>
 ***************************************************************************************************/
class Protocol:: Remed_Buddy:public Protocol::Message_Base{
public:
  virtual ~Remed_Buddy(){}
  bool parse(const Net_Message &m);
  std::string get_player_name() const { return player_name; }

  static std::string compose(const std::string &pn);
protected:
  std::string player_name;
};



/***************************************************************************************************
 * @class Buddies
 * Alist of buddies and their online/offline statust sent by the bot to a player upon request.
 * (<buddyName_1><+|->)*
 ***************************************************************************************************/
class Protocol:: Buddies:public Protocol::Message_Base{
public:
  virtual ~Buddies(){}
  bool parse(const Net_Message &m);

  static std::string compose(const Vector<std::string>& buddies, const Vector<uint1>& online);

  uint4 size() const  { return buddies.size(); }
  const std::string& get_name(uint4 i)  const { return buddies[i]; }
  const uint1   get_online(uint4 i) const { return online[i]; }

protected:

  Vector<std::string> buddies;
  Vector<uint1>  online;

};



/****************************************************************************************************
 * @class Player_List                                                                               
 * constains and parses the list of player in a game and the ready/unready status
 * (<playerName_1><+|->)*
 ****************************************************************************************************/
class Protocol:: Player_List:public Protocol::Message_Base{
public:

  virtual ~Player_List(){}
  bool   parse(const Net_Message &m);
  static std::string compose(const Vector<uint1>& present, const Vector<std::string>& players, const Vector<uint1>& ready);

  uint4 size() const  { return ready.size(); }
  const uint4     get_alias   (uint4 i) const { return  aliases[i]; }
  const std::string& get_name    (uint4 i) const { return  players[i]; }
  const uint1   get_ready   (uint4 i) const { return    ready[i]; }

protected:
  Vector<uint4>   aliases;
  Vector<std::string>  players;
  Vector<uint1>     ready;

};




/****************************************************************************************************
 * @class Who_List                                                                                  
 * constains and parses the list of all players currently online
 * (<buddyName_1>)*
 ****************************************************************************************************/
class Protocol:: Who_List:public Protocol::Message_Base{
public:

  virtual ~Who_List(){}
  bool   parse(const Net_Message &m);
  static std::string compose(const std::string& who);

  uint4 size() const  { return players.size(); }
  const std::string& get_name   (uint4 i) const { return  players[i]; }
  
protected:

  Vector<std::string>  players;
  
};



/***************************************************************************************************
 * @class GGS_Notify
 * composes a notify request and parses a GGS_Notify message
 * compose: <name> <+|>
 ***************************************************************************************************/
class Protocol::GGS_Notify : public Protocol::Message_Base{
public:
  bool parse(const Net_Message &m);

  /**
    returns the name of the player
    who just went on/off line
  */
  std::string get_name(){
    return name;
  }

  /**
    did the player go online or offline
  */
  bool get_online(){
    return online;
  }

  /**
    tells ggs to notify upon connection/disconnection
    of names
    @param names the names of users you wish to be notified about seperated by spaces
    @param notify true to add to notify list false to remove from notify list
  */
  static std::string compose(std::string names, bool notify);

protected:
  std::string name;
  bool   online;
};




/***************************************************************************************************
 * @class Register
 * sent by a new player to the bot to create an acount
 * <level> <name> <pass>
 ***************************************************************************************************/
class Protocol:: Register : public Protocol::Message_Base{
public:

  bool parse(const Net_Message &m);

  /**
    @return the name of the player
  */
  std::string get_name(){
    return name;
  }

  /**
    @return the name of the player
  */
  int get_level(){
    return level;
  }

  /**
    @return get the password
  */
  std::string get_pass(){
    return pass;
  }

  static std::string compose(int level, std::string names, std::string pass);

protected:
  int level;
  std::string name;
  std::string pass;
};




/***************************************************************************************************
 * @class Login
 * sent by a palyer to the bot to login
 * <level> <pass> <ip> <port>
 ***************************************************************************************************/
class Protocol:: Login : public Protocol::Message_Base{
public:

  bool parse(const Net_Message &m);

  /**
    @return the name of the player
  */
  int get_level() const{ return level;}
  
  /**
    @return get the password
  */
  std::string get_pass() const { return pass; }

  std::string get_ip() const { return ip; }
  int get_port()  const { return port; }
  /** @param port players dont need to specify a port */
  static std::string compose(int level, std::string pass, const std::string& ip, const sint4 port = 0);

protected:
  int level;
  std::string pass;
  std::string ip;
  int port;
};





/***************************************************************************************************
 * @class Down
 * sent by root to the bot to cleanly shut it down
 * no params
 ***************************************************************************************************/
class Protocol:: Down : public Protocol::Message_Base{
public:

  bool parse(const Net_Message &m);
  static std::string compose();

protected:

};




/***************************************************************************************************
 * @class Finger
 * parses a GGS finger message and composes a finger request
 * compose: <name>
 ***************************************************************************************************/
class Protocol:: Finger : public Protocol::Message_Base{
public:

  bool parse(const Net_Message &m);
  /**
    @param title the title of the info to retrive Ex (host, pass, login ...)
    @return the info corresponding to title, return std::string "NULL" if not found
  */
  std::string get(const std::string& title);
  static std::string compose(const std::string& name);

protected:
  typedef __gnu_cxx::hash_map<std::string, std::string, StringHash> InfoList;
  InfoList list;

};


/***************************************************************************************************
 * @class List
 * used to request a list from the bot
 * Ex: list buddies, list who, list invitations
 * <list_name>
 ***************************************************************************************************/
class Protocol:: List : public Protocol::Message_Base{
public:

  bool parse(const Net_Message &m);
  static std::string compose(const std::string& listName);

  const std::string& get_list_name(){return name;}

protected:
  static bool validName(const std::string& name_);
  std::string name;

};

/***************************************************************************************************
 * @class Worker_Available
 * sent by a worker to the bot upon finishing a game, worker is ready to go again
 * <ip>
 ***************************************************************************************************/
class Protocol:: Worker_Available : public Protocol::Message_Base{
public:

  bool parse(const Net_Message &m);
  std::string get_ip() const { return ip; }
  static std::string compose(const std::string& ip);

protected:
  std::string ip;
};

#endif
