//-----------------------------------------------------------------------------
/** @file bucketfringesearch.h
    FringeSearch using multiple buckets.

    $Id: bucketfringesearch.h,v 1.1 2006/12/31 04:09:41 orts_furtak Exp $
    $Source: /usr/bodo1/cvs/cvs/orts3/libs/pathfinding/ua_path_lib/src/bucketfringesearch.h,v $
*/
//-----------------------------------------------------------------------------

#ifndef PATHFIND_BUCKETFRINGESEARCH_H
#define PATHFIND_BUCKETFRINGESEARCH_H

#include <limits>
#include "marker.h"
#include "search.h"
#include "util.h"

//-----------------------------------------------------------------------------

namespace PathFind
{
    using namespace std;

    /** Statically allocated set of linked lists used in BucketFringeSearch. */
    class BucketFringe
    {
    public:
        BucketFringe(int numberBuckets);
        
        int getCurrentNode() const
        {
            return m_currentNode;
        }
        
        void init(int numberNodes);
        
        /** Insert node after current node. */
        void insertNode(int nodeId);
        
        /** Insert node at front in bucket. */
        void insertNode(int nodeId, int bucket);

        bool isEmpty(int bucket)
        {
            return (getHeadNode(bucket) == NO_NODE);
        }
        
        void nextNode();
        
        void print(ostream& o) const;
        
        void removeCurrentNode()
        {
            removeNode(m_currentNode);
        }
        
        void startIteration(int bucket);
        
    private:
        /** Node used in BucketFringe. */
        class Node
        {
        public:
            int m_next;
            
            int m_previous;
        };
        

        int m_currentBucket;

        int m_currentNode;
        
        int m_numberBuckets;

        int m_numberNodes;
        
        vector<int> m_headNode;
        
        /** Stores bucket number for each node in the environment.
            Contains -1 for nodes not in any bucket.
        */
        vector<int> m_nodeInBucket;
        
        vector<Node> m_nodes;
        
        
        bool allLowerBucketsEmpty(int bucket) const;

        int getHeadNode(int bucket) const
        {
            assert(bucket >= 0);
            assert(bucket < m_numberBuckets);
            return m_headNode[bucket];
        }

        Node& getNode(int nodeId)
        {
            assert(nodeId >= 0 && nodeId < m_numberNodes);
            return m_nodes[nodeId];
        }
        
        const Node& getNode(int nodeId) const
        {
            assert(nodeId >= 0 && nodeId < m_numberNodes);
            return m_nodes[nodeId];
        }
        
        void removeNode(int nodeId);

        void setHeadNode(int bucket, int nodeId)
        {
            assert(bucket >= 0);
            assert(bucket < m_numberBuckets);
            m_headNode[bucket] = nodeId;
        }
    };

    /** FringeSearch using multiple buckets.

        %Statistics (see Search::createStatistics and Search::getStatistics):
        - "cpu_time"
        - "iterations"
        - "nodes_expanded"
        - "nodes_visited"
        - "path_cost"
        - "path_length"
    */
    template<class MARKER = MarkerFastClear>
    class BucketFringeSearch
        : public Search
    {
    public:
        BucketFringeSearch(int numberBuckets, int deltaBucketFValue);

        StatisticsCollection createStatistics();

        bool findPath(const Environment& env, int start, int target);

        const vector<int>& getPath() const
        {
            return m_path;
        }

        const StatisticsCollection& getStatistics() const;

        const vector<char>& getVisitedNodes() const
        {
            return m_visitedNodes;
        }

    private:   
        /** Cache element for g-value used in BucketFringeSearch. */
        class CacheElement
        {
        public:
            int m_gValue;
            
            int m_iteration;

            int m_parent;
        };


        static const bool BUCKETFRINGE_DEBUG = false;
        
        int m_deltaBucketFValue;

        int m_nodesExpanded;

        int m_nodesVisited;

        int m_numberBuckets;

        int m_startBucketFValue;

        int m_target;

        const Environment* m_env;

        BucketFringe m_fringe;

        vector<CacheElement> m_cache;
        
        MARKER m_isCacheValid;
        
        vector<char> m_visitedNodes;

        vector<int> m_fMinInBucket;

        vector<int> m_path;

        StatisticsCollection m_statistics;


        void addCache(int nodeId, int gValue, int iteration, int parent);

        bool checkCache(int nodeId, int gValue, int iteration) const;

        void constructPath(int start);

        int findNextBucket(int bucket);

        int getBucket(int fValue);

        void getFromCache(int nodeId, int& gValue, int& parent) const;

        bool doIteration(int iteration, int currentBucket, int fLimit,
                         int& fMin);

        void init(int numberNodes);

        void markVisitedNode(int nodeId, int iteration);

        void updatefMinInBucket(int bucket, int fValue);
    };
}

//-----------------------------------------------------------------------------

namespace PathFind
{
    using namespace std;

    template<class MARKER>
    BucketFringeSearch<MARKER>::BucketFringeSearch(int numberBuckets,
                                                   int deltaBucketFValue)
        : m_fringe(numberBuckets),
          m_fMinInBucket(numberBuckets, numeric_limits<int>::max()),
          m_statistics(createStatistics())
    {
        assert(deltaBucketFValue > 0);
        m_numberBuckets = numberBuckets;
        m_deltaBucketFValue = deltaBucketFValue;
    }
    
    template<class MARKER>
    void BucketFringeSearch<MARKER>::addCache(int nodeId, int gValue,
                                              int iteration, int parent)
    {
        CacheElement& cacheElement = m_cache[nodeId];
        cacheElement.m_gValue = gValue;
        cacheElement.m_iteration = iteration;
        cacheElement.m_parent = parent;
        m_isCacheValid.set(nodeId);
    }

    template<class MARKER>
    bool BucketFringeSearch<MARKER>::checkCache(int nodeId, int gValue,
                                                int iteration) const
    {
        if (! m_isCacheValid.get(nodeId))
            return false;
        const CacheElement& cacheElement = m_cache[nodeId];
         int gValueCached = cacheElement.m_gValue;
        return (gValue > gValueCached
                || (gValue == gValueCached
                    && iteration == cacheElement.m_iteration));
    }

    template<class MARKER>
    void BucketFringeSearch<MARKER>::constructPath(int start)
    {
        m_path.clear();
        if (! m_isCacheValid.get(m_target))
            return;
        int nodeId = m_target;
        do
        {
            m_path.push_back(nodeId);
            nodeId = m_cache[nodeId].m_parent;
            assert(m_isCacheValid.get(nodeId));
        }
        while (nodeId != start);
        m_path.push_back(nodeId);
    }

    template<class MARKER>
    StatisticsCollection BucketFringeSearch<MARKER>::createStatistics()
    {
        StatisticsCollection collection;
        collection.create("cpu_time");
        collection.create("iterations");
        collection.create("nodes_expanded");
        collection.create("nodes_visited");
        collection.create("path_cost");
        collection.create("path_length");
        return collection;
    }

    template<class MARKER>
    int BucketFringeSearch<MARKER>::findNextBucket(int bucket)
    {
        assert(bucket >= 0);
        assert(bucket < m_numberBuckets);
        for (; bucket < m_numberBuckets; ++bucket)
            if (! m_fringe.isEmpty(bucket))
                return bucket;
        return m_numberBuckets - 1;
    }

    template<class MARKER>
    int BucketFringeSearch<MARKER>::getBucket(int fValue)
    {
        int diffFValue = fValue - m_startBucketFValue;
        if (diffFValue <= 0)
            return 0;
        int bucket = diffFValue / m_deltaBucketFValue;
        if (bucket >= m_numberBuckets)
            bucket = m_numberBuckets - 1;
        return bucket;
    }

    template<class MARKER>
    bool BucketFringeSearch<MARKER>::doIteration(int iteration,
                                                 int currentBucket, int fLimit,
                                                 int& fMin)
    {
        if (BUCKETFRINGE_DEBUG)
            cerr << "Current bucket = " << currentBucket << '\n';
        m_fringe.startIteration(currentBucket);
        vector<Environment::Successor> successors;
        while (true)
        {
            int nodeId = m_fringe.getCurrentNode();
            if (BUCKETFRINGE_DEBUG)
                cerr << "Visit " << nodeId << '\n';
            if (nodeId == NO_NODE)
                return false;
            ++m_nodesVisited;
            markVisitedNode(nodeId, iteration);
            int gValue;
            int lastNode;
            getFromCache(nodeId, gValue, lastNode);
            int fValue = gValue + m_env->getHeuristic(nodeId, m_target);
            if (fValue > fLimit)
            {
                if (fValue < fMin)
                    fMin = fValue;
                m_fringe.nextNode();
                continue;
            }
            if (nodeId == m_target)
            {
                m_statistics.get("path_cost").add(gValue);
                return true;
            }
            addCache(nodeId, gValue, iteration, lastNode);
            m_env->getSuccessors(nodeId, lastNode, successors);
            int numberSuccessors = successors.size();
            ++m_nodesExpanded;
            for (int i = numberSuccessors; i--; )
            {
                const Environment::Successor& successor = successors[i];
                int targetNodeId = successor.m_target;
                int targetGValue = gValue + successor.m_cost;
                if (checkCache(targetNodeId, targetGValue, iteration))
                    continue;
                int targetFValue =
                    targetGValue + m_env->getHeuristic(targetNodeId, m_target);
                if (targetFValue > fLimit)
                {
                    if (targetFValue < fMin)
                        fMin = targetFValue;
                }
                int bucket = getBucket(targetFValue);
                if (BUCKETFRINGE_DEBUG)
                    cerr << "Insert " << targetNodeId << '(' << targetFValue
                         << ") into " << bucket << '\n';
                if (bucket == currentBucket)
                    m_fringe.insertNode(targetNodeId);
                else
                {
                    m_fringe.insertNode(targetNodeId, bucket);
                    updatefMinInBucket(bucket, targetFValue);
                }
                addCache(targetNodeId, targetGValue, iteration, nodeId);
            }
            m_fringe.removeCurrentNode();
            m_fringe.nextNode();
        }
    }

    template<class MARKER>
    bool BucketFringeSearch<MARKER>::findPath(const Environment& env,
                                              int start, int target)
    {
        assert(env.isValidNodeId(start));
        assert(env.isValidNodeId(target));
        Clock clock;
        m_env = &env;
        m_target = target;
        int numberNodes = env.getNumberNodes();
        init(numberNodes);
        m_nodesExpanded = 0;
        m_nodesVisited = 0;
        int iteration = 1;
        int fLimit = env.getHeuristic(start, target);
        m_startBucketFValue = fLimit;
        int bucket = getBucket(fLimit);
        m_fringe.insertNode(start, bucket);
        addCache(start, 0, 0, NO_NODE);
        while (true)
        {
            if (BUCKETFRINGE_DEBUG)
                cerr << "Iteration " << iteration << ":\n\nlimit = "
                     << fLimit << '\n';
            int fMin = numeric_limits<int>::max();
            if (doIteration(iteration, bucket, fLimit, fMin))
                break;
            if (m_fringe.isEmpty(bucket))
            {
                bucket = findNextBucket(bucket);
                fLimit = m_fMinInBucket[bucket];
                if (BUCKETFRINGE_DEBUG)
                    cerr << "Empty bucket/limit: " 
                         << bucket << " " << fLimit << '\n';
            }
            else
            {
                fLimit = fMin;
                bucket = getBucket(fLimit);
                if (BUCKETFRINGE_DEBUG)
                    cerr << "Non-empty bucket/limit: " 
                         << bucket << " " << fLimit << '\n';
            }
            assert(fLimit != numeric_limits<int>::max());
            ++iteration;
            if (BUCKETFRINGE_DEBUG)
            {
                m_fringe.print(cerr);
                cerr << '\n';
            }
        }
        constructPath(start);
        m_statistics.get("iterations").add(iteration);
        m_statistics.get("nodes_visited").add(m_nodesVisited);
        m_statistics.get("nodes_expanded").add(m_nodesExpanded);
        m_statistics.get("path_length").add(m_path.size());
        m_statistics.get("cpu_time").add(clock.get());
        return true;
    }

    template<class MARKER>
    void BucketFringeSearch<MARKER>::getFromCache(int nodeId, int& gValue,
                                                  int& parent) const
    {
        assert(m_isCacheValid.get(nodeId));
        const CacheElement& cacheElement = m_cache[nodeId];
        parent = cacheElement.m_parent;
        gValue = cacheElement.m_gValue;
    }

    template<class MARKER>
    const StatisticsCollection& BucketFringeSearch<MARKER>::getStatistics()
        const
    {
        return m_statistics;
    }

    template<class MARKER>
    void BucketFringeSearch<MARKER>::init(int numberNodes)
    {
        m_visitedNodes.resize(numberNodes);
        fill(m_visitedNodes.begin(), m_visitedNodes.end(), ' ');
        m_isCacheValid.init(numberNodes);
        m_cache.resize(numberNodes);
        m_fringe.init(numberNodes);
        m_statistics.clear();
        fill(m_fMinInBucket.begin(), m_fMinInBucket.end(),
             numeric_limits<int>::max());
    }

    template<class MARKER>
    void BucketFringeSearch<MARKER>::markVisitedNode(int nodeId, int iteration)
    {
        if (m_visitedNodes[nodeId] != ' ')
            return;
        m_visitedNodes[nodeId] = getVisitedNodeLabel(iteration);
    }

    template<class MARKER>
    void BucketFringeSearch<MARKER>::updatefMinInBucket(int bucket,
                                                        int fValue)
    {
        assert(bucket >= 0);
        assert(bucket < m_numberBuckets);
        if (fValue < m_fMinInBucket[bucket])
            m_fMinInBucket[bucket] = fValue;
    }
}

//-----------------------------------------------------------------------------

#endif
