/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "RTS_Game.H"
#include "Element.H"
#include <algorithm>

// fixme: speed improvement (pythagoras) possible!

// for debugging ...
static sint4 counter = 0;
const sint4 COUNT = 7060;

#if 0
static sint8 llabs(sint8 &x) {
  if (x >= 0) return x;
  else return -x;
}
#endif

static sint8 sgn(sint8 &x) {
  if (x >= 0) return 1LL;
  return -1LL;
}

static sint4 sgn(sint4 &x) {
  if (x >= 0) return 1;
  return -1;
}

bool RTS_Game::check_intersections(sint4 &ind1, sint4 &ind2)
{
  uint4 n = elems.size();

  FORU (i, n) {
    Element &e1 = *elems[i];
    Pos p1(e1.sv_pos_x, e1.sv_pos_y);
    Scalar r1 = e1.sv_radius;
    for (uint4 j = i+1; j < n; j++) {
      Element &e2 = *elems[j];
      Pos p2(e2.sv_pos_x, e2.sv_pos_y);
      Scalar r2 = e2.sv_radius;
      Pos delta = p1;
      delta.sub(p2);
      Scalar r = FixPoint::add(r1, r2) - 3; // slack
      if (delta.max_norm() >= r) continue;
      if (delta.l2_norm() < r) {
	ind1 = i; ind2 = j;
	cout << "delta=" << (delta.l2_norm() - r) << endl;
	return true;
      }
    }
  }
  return false;
}


// compute collision time
// return >= 2 if no such collision exists
// return -1 if too close

real8 RTS_Game::collision_time2(Element &e1, Element &e2, real8 ct)
{
  if (ct < 0.0L) { cout << "########### ct<0: " << ct << endl; ct = 0.0L; }
  if (ct > 1.0L) { cout << "########### ct>1: " << ct << endl; ct = 1.0L; }
  
  counter++;

#if TEST_OUT
  
  cout << "COUNTER= " << counter << endl;

  if (counter == COUNT) {
    cout << "HERE I AM" << endl;
  }

  cout << "e1:  p="; e1.mv.sg_pos.write();
  cout << " h="; e1.sv_heading.write();
  cout << " m="; cout << e1.sv_moving;
  cout << endl;

  cout << "e2:  p="; e2.mv.sg_pos.write();
  cout << " h="; e2.sv_heading.write();
  cout << " m="; cout << e2.sv_moving;
  cout << endl;

#endif
  
  // 2D for now

  Pos dh(0,0), dp(0,0);

  if (e1.sv_moving) {

    dh = e1.sv_fine_heading;
    dp = e1.mv.fine_pos;
    
  } else {

    dp.x = fine_grid(e1.sv_pos_x);
    dp.y = fine_grid(e1.sv_pos_y);    

  }

  if (e2.sv_moving) {

    dh.sub(e2.sv_fine_heading);
    dp.sub(e2.mv.fine_pos);
    
  } else {

    Pos p2;
    p2.x = fine_grid(e2.sv_pos_x);
    p2.y = fine_grid(e2.sv_pos_y);    
    dp.sub(p2);
  }

  Pos moved = dh;
  moved.scale(ct);
  dp.add(moved);
  
  if (dh.x == 0 && dh.y == 0) return 2.0; // no relative position change
  
  sint8 p = sint8(dp.x) * sint8(dp.x) + sint8(dp.y) * sint8(dp.y);
  sint8 h = sint8(dh.x) * sint8(dh.x) + sint8(dh.y) * sint8(dh.y);

  // inner product, direction angle
  sint8 g = sint8(dh.x) * sint8(dp.x) + sint8(dh.y) * sint8(dp.y);

#if TEST_OUT
  cout << "dh="; dh.write();
  cout << " dp="; dp.write();
  cout << " g=" << g
       << " RES=" << sub_grid_res << endl;
#endif

  sint8 fr = (e1.sv_radius + e2.sv_radius) << sub_grid_res_bits;
  sint8 d = square(fr) - p;

  if (g >= 0LL) { // objects are moving away from each other

    if (d >= 0LL) { // objects intersect or touch each other

      // eight main directions => objects don't come closer (even when pos is rounded)

      if (dh.x == 0 || dh.y == 0 || abs(dh.x) == abs(dh.y)) return 2.0;

#if TEST_OUT      
      cout << "++++++++++++++++++++++++++" << endl;
#endif
      
      // directions must be opposing! otherwise they can get closer after
      // rounding to grid points!

      Pos h1(0,0), h2(0,0);

      if (e1.sv_moving) {
	h1 = e1.sv_fine_heading;
      }
      if (e2.sv_moving) {
	h2 = e2.sv_fine_heading;
      }
      
      if (h1.x != 0 && h2.x != 0 && sgn(h1.x) == sgn(h2.x)) return 0.0;
      if (h1.y != 0 && h2.y != 0 && sgn(h1.y) == sgn(h2.y)) return 0.0;      

#if TEST_OUT      
      cout << "/////////////////////" << endl;      
#endif
      
      // opposing directions => objects don't get closer
      return 2.0;
        
    } else return 2.0;  

  }
  
#if TEST_OUT  
  cout << "d= " << d << " p=" << p
       << " " << ((sint8(e1.sv_radius + e2.sv_radius) << sub_grid_res_bits)) << endl;
#endif
  
  if (d >= 0LL) return 0.0;  // intersection or touch => collision
  
  // solving quadratic equation
  
  // 2*t*g + t^2*h = d   ->  a*t^2 + b*t + c = 0;

  sint8 a = h;
  sint8 b = 2LL*g;
  sint8 c = -d;

  sint8 MAX = 1LL << 50;
  if (std::llabs(a) > MAX) ERR("g out of range");
  if (std::llabs(b) > MAX) ERR("d out of range");
  if (std::llabs(c) > MAX) ERR("h out of range");  


  real8 inv_b = 1.0/real8(b);
  real8 a_div_b = real8(a) * inv_b;
  real8 c_div_b = real8(c) * inv_b;

#if TEST_OUT  
  cout << "OOO " << a << " " << b << " " << c << endl;
  cout << a_div_b << " " << c_div_b << endl;
#endif
  
  // b^2-4ac < 0?

  real8 prod = a_div_b * c_div_b * 4.0L;
  
  if (prod > 1.0L) {
    return 3.0;  // no collision
  }

  real8 discr = 1.0L - prod;
  if (discr < 0) ERR("discr < 0");
  discr = std::llabs(b) * sqrt(discr);

  // robust root computation (Numerical Recipies)
  
  real8 q = -0.5L * (b + sgn(b) * discr);
  real8 t1 = q / a;
  real8 t2 = c / q;

#if TEST_OUT
  cout << "prod= " << prod << endl;
  cout << "t1=" << t1 << endl;
  cout << "t2=" << t2 << endl;  
  cout << "v1=" << (a*t1*t1+b*t1+c) << endl;
  cout << "v2=" << (a*t2*t2+b*t2+c) << endl;  
#endif
  
  if (t2 < t1) { real8 t = t1; t1 = t2; t2 = t; }

  if (t1 < 0.0L) t1 = t2;
  if (t1 < 0.0L) return 4.0;
  if (t1 > 1.0L) {
    // cout << "T1=" << t1 << endl;
    return 5.0;
  }

  return t1;
}


bool RTS_Game::visit(const vector<Element*> &elems, sint4 i, sint4 *&comp)
{
  visit_n++;
  Element &e = *elems[i];
  
  if (e.visited) return false;

  e.visited = true;
  bool mobile = false;

  *comp++ = i;

  if (e.sv_moving) mobile = true;

  const NeighborInfo &ne = e.neighbors;
  
  FORS (ni, ne.size()) mobile |= visit(elems, ne[ni], comp);

  return mobile;
}
