/*
    (c) Michael Buro, mic@research.nj.nec.com
    NEC Research Institute
    4 Independence Way
    Princeton, NJ 08540, USA

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "RTS_Game.H"
#include "Element.H"
#include "CycleCounter.H"
#include "PS_Tree.H"

static uint8 t_n = 0;
static uint8 t_sum = 0;

static uint4 ccc = 0;

struct PointInfo {
  sint4 y;
  sint4 xmin, xmax;
  sint2 ind;
  sint1 is_max;
  sint1 sight;  

  void write(ostream &os = cout) const {
    os << "y=" << y
       << " xmin=" << xmin
       << " xmax=" << xmax
       << " ind=" << ind
       << " max=" << sint4(is_max)
       << " sight=" << sint4(sight);
  }
};

struct PointInfoCmp {

  bool operator()(const PointInfo &p1, const PointInfo &p2) {
    sint4 d = p1.y - p2.y;
    if (d < 0) return true;
    if (d > 0) return false;
    d = p1.xmin - p2.xmin;
    if (d < 0) return true;
    if (d > 0) return false;
    d = p1.xmax - p2.xmax;
    return d < 0;
  }

};


static sint4 intlog2(sint4 x) {

  if (x <= 0) ERR("x <= 0 in intlog2");
  sint4 i = 0, m = 1;
  while (m < x) { i++; m <<= 1; }
  return i;
}

void RTS_Game::compute_visible_elements()
{
  CycleCounter start_time;

  ccc = 0;

  const uint4 n = elems.size();
  sint4 dimx, dimy;
  opt.get("-x", dimx);
  opt.get("-y", dimy);

  if (n <= 0) return;

  const sint4 n2log2 = intlog2(2*n);
  const sint4 n2ceil2 = 1 << n2log2 ;               // 2n different ids (unique x-values)
  const sint4 ub = n2ceil2*::max(dimx, dimy) + 2*n; // upper bound for all PST coordinates
  PS_Tree<sint4> psts[2];

  psts[0].set_ub(ub);
  psts[1].set_ub(ub);  

  vector<PointInfo> scan_vec;
  scan_vec.reserve(4*n);

  if (n >= 32768) ERR("n too large");  // PointInfo.ind = sint2
  
  FORU (i, n) {

    Element &e = *elems[i];
    e.neighbors.set_size(0);
    e.prev_seen_by = e.seen_by;
    e.seen_by.clear();
    e.seen_by.insert(e.sv_owner);
    sint4 x  = e.sv_pos_x;
    sint4 y  = e.sv_pos_y;
    sint4 ro = e.sv_radius;
    sint4 rs = e.sv_sight_range;

    if (ro <= 0 || rs <= 0) ERR("radius/sight <= 0");
    
    PointInfo pi;
    
    // object bounding rectangle
    
    pi.ind = i;

    pi.sight  = 0;
    pi.is_max = false;
    pi.y    = ::max(0, y - ro);
    pi.xmin = (::max(0,    x - ro) << n2log2);
    pi.xmax = (::min(dimx, x + ro) << n2log2) + 2*i;

    scan_vec.push_back(pi);

    pi.is_max = true;
    pi.y = ::min(dimy, y + ro);

    scan_vec.push_back(pi);
    
    // sight bounding rectangle

    pi.sight = 1;
    pi.is_max = false;
    pi.y    = ::max(0, y - rs);
    pi.xmin = (::max(0,    x - rs) << n2log2);
    pi.xmax = (::min(dimx, x + rs) << n2log2) + 2*i+1;

    scan_vec.push_back(pi);

    pi.is_max = true;
    pi.y = ::min(dimy, y + rs);

    scan_vec.push_back(pi);
  }

  sort(scan_vec.begin(), scan_vec.end(), PointInfoCmp());
  
  sint4 sn = 4*n;
  assert(sint4(scan_vec.size()) == sn);

  vector<Pair<sint4>*> out;
  out.reserve(200);

  vector<Node<sint4> > nodes(sn);
  
  FORS (i, sn) {

    PointInfo &pi = scan_vec[i];

    Pair<sint4> pair;
    pair.x    = pi.xmax;  // exchange x-y for interval intersection
    pair.y    = pi.xmin;
    pair.data = pi.ind;

    //    cout << "current pinfo: "; pi.write(cout); cout << endl;
    
    if (pi.is_max) {

      if (!psts[pi.sight].remove_pair(pair)) ERR("delete failed");

    } else {

      // new rectangle appeared

      nodes[i].pair = pair;
      psts[pi.sight].insert_node(&nodes[i]);
      
      psts[1-pi.sight].enumerate_rectangles(pi.xmin, ub, pi.xmax, out);

      Element &e1 = *elems[pi.ind];
      
      if (pi.sight) {

	// new sight rectangle

	//cout << "NEW SIGHT " << out.size() << endl;

	FORALL (vector<Pair<sint4>*>, out, i) {
	  Element &e2 = *elems[(*i)->data];
	  sint4 r2 = e1.sv_sight_range + e2.sv_radius;
	  assert(r2 <= 32768);
	  assert(abs(e1.sv_pos_x-e2.sv_pos_x) <= r2);
	  assert(abs(e1.sv_pos_y-e2.sv_pos_y) <= r2);	  
	  if (&e1 != &e2 &&
         square(e1.sv_pos_x-e2.sv_pos_x)+square(e1.sv_pos_y-e2.sv_pos_y) <= r2*r2) {
	    // cout << pi.ind << "->" << ((*i)->data) << endl;

	    // e1 sees e2
	    e1.neighbors.push_back((*i)->data);
	    e2.seen_by.insert(e1.sv_owner);
	  }
	}
	
      } else {

	// new object rectangle

	//cout << "NEW OBJ " << out.size() << endl;
	
	FORALL (vector<Pair<sint4>*>, out, i) {
	  Element &e2 = *elems[(*i)->data];
	  sint4 r2 = e1.sv_radius + e2.sv_sight_range;
	  assert(r2 <= 32768);
	  assert(abs(e1.sv_pos_x-e2.sv_pos_x) <= r2);
	  assert(abs(e1.sv_pos_y-e2.sv_pos_y) <= r2);
	  if (&e1 != &e2 &&
	      square(e1.sv_pos_x-e2.sv_pos_x)+square(e1.sv_pos_y-e2.sv_pos_y) <= r2*r2) { 
	    e2.neighbors.push_back(sint4(pi.ind));
	    e1.seen_by.insert(e2.sv_owner);
	  }
	}
      }
    }
  }

  assert(psts[0].size() == 0 && psts[1].size() == 0);
  
  CycleCounter end_time;

  bool stat;
  opt.get("-stat", stat);

  if (stat) {
    t_n++;
    sint8 delta = end_time.count()-start_time.count();
    t_sum += delta;
    
    form(cout, "visibility:  %10lld %lld\n", t_sum/t_n, delta);
  }
}
